from .BaseReader import BaseReader
from .BaseWriter import BaseWriter
from .AscReader import AscReader
from .iGPReader import iGPReader
from .CsvWriter import CsvWriter
from .BoreholeReader import BoreholeReader
from .PflotranExplicitWriter import PflotranExplicitWriter
from .PflotranImplicitWriter import PflotranImplicitWriter




"""
Contains general purpose utility functions
"""
import yaml
import pydelling.readers as readers
import pydelling.interpolation as interpolation
import numpy as np
from pathlib import Path
import os
try:
    from pydelling.paraview_processor.filters import BaseFilter, PlotOverLineFilter
except:
    from pydelling.paraview_processor.filters import BaseFilter
import pandas as pd
from box import Box
from scipy.stats import lognorm


def interpolate_permeability_anisotropic(perm_filename, mesh_filename=None, mesh=None):
    perm = readers.CentroidReader(filename=perm_filename, header=True)
    if mesh is None:
        assert mesh_filename is not None, "A mesh file needs to be given under the mesh_filename tag"
        mesh = readers.CentroidReader(filename=mesh_filename, header=False)
    else:
        mesh = mesh
    interpolator = interpolation.SparseDataInterpolator(interpolation_data=perm.get_data(),
                                          mesh_data=mesh.get_data())
    interpolator.interpolate(method="nearest")
    if mesh is None:
        return interpolator, mesh
    else:
        return interpolator


def interpolate_centroid_to_structured_grid(centroid: np.ndarray,
                                            var: np.ndarray,
                                            automatic=True,
                                            n_x=100,
                                            n_y=100,
                                            origin_x=0.0,
                                            final_x=1.0,
                                            origin_y=0.0,
                                            final_y=1.0) -> np.ndarray:
    """Reads a set of 2D centroid points and interpolates them into a structured grid"""
    # Check that the centroid file is a 2D array
    assert centroid.shape[1] == 2, "The given centroid file is not a 2D array"
    _var = var
    _centroid = centroid
    _dx = abs(final_x - origin_x)/n_x
    _dy = abs(final_y - origin_y) / n_y
    linspace_x = np.linspace(origin_x, final_x, n_x)
    linspace_y = np.linspace(origin_y, final_y, n_y)
    # reshape private variable _var to shape [:, 1]
    if len(var.shape) == 1:
        _var = np.reshape(_var, (var.shape[0], 1))
    grid_x, grid_y = np.meshgrid(linspace_x, linspace_y)


def aperture_from_a_xy_point_old(dataset: BaseFilter, x_point, y_point, line_interpolator, variable=None, target_value=1.0, threshold=0.45, line_resolution=100):
    """
    This method computes the aperture at a given position in the XY plane.
    Args:
        x_point: X coordinate
        y_point: Y coordinate
        variable: The variable to consider that indicates the aperture
        target_value: Value of the variable when the fracture is open
        threshold: Value of the maximum variation from the target_value |line[variable] - target_value| < threshold is assumed
        line_resolution: Resolution of the line interpolation that is being used
    Returns:
        The value of the aperture at a given point
    """
    point_1 = [x_point, y_point, dataset.z_min]
    point_2 = [x_point, y_point, dataset.z_max]
    line_interpolator.set_points(point_1=point_1, point_2=point_2)
    line_interpolator.set_line_resolution(line_resolution)

    line_interpolation_point_data = line_interpolator.point_data
    dataset_mesh_points = dataset.mesh_points
    if variable:
        # If a target variable is defined,
        # print(abs(line_interpolation_point_data[variable] - target_value))
        line_interpolation = line_interpolation_point_data[abs(line_interpolation_point_data[variable] - target_value) < threshold]
        z_points: pd.Series = dataset_mesh_points.iloc[line_interpolation.index]["z"]
        if len(z_points) > 0:
            # aperture = abs(z_points.max() - z_points.min())
            aperture = z_points.max()
        else:
            aperture = 0.0
        return aperture

    if not variable:
        # Calculate directly the aperture
        z_points: pd.Series = dataset_mesh_points.iloc[line_interpolation_point_data.index]["z"]
        if len(z_points) > 0:
            aperture = abs(z_points.max() - z_points.min())
        else:
            aperture = 0.0
        return aperture

def aperture_from_a_xy_point(dataset: BaseFilter, x_point, y_point, line_interpolator, variable=None, target_value=1.0, threshold=0.45, line_resolution=100, method='explicit'):
    """
    This method computes the aperture at a given position in the XY plane.
    Args:
        x_point: X coordinate
        y_point: Y coordinate
        variable: The variable to consider that indicates the aperture
        target_value: Value of the variable when the fracture is open
        threshold: Value of the maximum variation from the target_value |line[variable] - target_value| < threshold is assumed
        line_resolution: Resolution of the line interpolation that is being used
        method: Method to use to calculate the aperture. It can be 'explicit' or 'implicit'
    Returns:
        The value of the aperture at a given point
    """
    point_1 = [x_point, y_point, dataset.z_min]
    point_2 = [x_point, y_point, dataset.z_max]
    line_interpolator.set_points(point_1=point_1, point_2=point_2)
    line_interpolator.set_line_resolution(line_resolution)
    line_interpolator_resolution = (dataset.z_max - dataset.z_min) / line_resolution


    line_interpolation_point_data = line_interpolator.point_data
    dataset_mesh_points = dataset.mesh_points
    if variable:
        # If a target variable is defined,
        # print(abs(line_interpolation_point_data[variable] - target_value))
        # line_interpolation = line_interpolation_point_data[abs(line_interpolation_point_data[variable] - target_value) < threshold]
        # z_points: pd.Series = dataset_mesh_points.iloc[line_interpolation.index]["z"]
        if method == 'explicit':
            line_interpolation = line_interpolation_point_data[abs(line_interpolation_point_data[variable] - target_value) < threshold]
            z_points: pd.Series = dataset_mesh_points.iloc[line_interpolation.index]["z"]
            aperture = len(z_points) * line_interpolator_resolution
            return aperture
        elif method == 'implicit':
            eps_field = (line_interpolation_point_data[variable] - line_interpolation_point_data[variable].min())
            eps_field = eps_field / eps_field.max()
            aperture = (eps_field * line_interpolator_resolution).sum()
            return aperture

    if not variable:
        # Calculate directly the aperture
        z_points: pd.Series = dataset_mesh_points.iloc[line_interpolation_point_data.index]["z"]
        if len(z_points) > 0:
            aperture = abs(z_points.max() - z_points.min())
        else:
            aperture = 0.0
        return aperture


def get_root_path() -> Path:
    """Returns path to the root of the project"""
    return Path(__file__).parent


def get_config_path() -> Path:
    """Returns path to the root of the project"""
    return Path(__file__).parent / "config"


def test_data_path() -> Path:
    """Returns path to the root of the project"""
    return Path(__file__).parent.parent / "tests/test_data"

def runtime_path():
    return Path(os.getcwd())


def read_local_config():
    def read_config(config_file: Path = "./config.yaml"):
        """
        Reads the configuration file
        :param config_file:
        :return:
        """
        with open(config_file) as file:
            context = yaml.load(file, Loader=yaml.FullLoader)
        return Box(context, default_box=True)

    _config_file = list(
        Path(os.getcwd()).glob("**/*config.yml") and Path(os.getcwd()).glob("**/*config.yaml") and Path().cwd().glob(
            "*config*.yml") and Path().cwd().glob(
            "*config*.yaml"))
    _config_file = _config_file if _config_file else list(Path(__file__).parent.glob("config.yml"))
    assert len(_config_file) == 1, "Please provide a configuration file that has a '*config.yaml' name structure"
    config = read_config(config_file=_config_file[0])
    return config


def sample_values_from_dict(input_dict: dict, n: int, write_to_file=True, return_generator=False) -> list:
    """
    This method reads a dictionary and generates n samples from them based on the
    following criteria.

    The structure of the dictionary should be as follows:

    name_of_dict:
        material_name_i:
            type: type of the desired distribution
            option_keys: these keys depend on the chosen distribution type
        material_name_i:

    Type of implemented distributions:
        constant[value]: sets a constant value for the variable
        log_normal[mean, std]: generates a log-normal distribution based on the mean, std values
        specified in natural scale.

    Args:
        write_to_file: optional argument that writes a csv file with the generated cases. Defaults to True.
        d: dictionary to extract the samples from
        n: integer specifying the number of generated samples
    Returns:
        A dictionary containing the sampled results
    """
    class BaseDistribution:
        def __init__(self):
            pass
        def run(self) -> float:
            """
            This method generates a value of the given distribution
            Returns: a float number containing the value
            """

    class ConstantDistribution(BaseDistribution):
        def __init__(self, value):
            super().__init__()
            self.value = value

        def run(self):
            return self.value

    class NormalDistribution(BaseDistribution):
        def __init__(self, mean, std, log=False):
            super().__init__()
            self.mean = float(mean)
            self.std = float(std)
            self.log = log

        def run(self):
            if not self.log:
                return np.random.normal(self.mean, self.std)
            else:
                return np.random.lognormal(self.mean, self.std)

    # Process each material and create sample generators
    generator_dict = {}
    for material in input_dict:
        material_dict = input_dict[material]
        if material_dict['type'] == 'constant':
            generator_dict[material] = ConstantDistribution(value=material_dict['value'])
        elif material_dict['type'] == 'normal':
            generator_dict[material] = NormalDistribution(mean=material_dict['mean'],
                                                          std=material_dict['std'],
                                                          )
        elif material_dict['type'] == 'log-normal' or material_dict['type'] == 'log_normal':
            generator_dict[material] = NormalDistribution(mean=material_dict['mean'],
                                                          std=material_dict['std'],
                                                          log=True,
                                                          )
        elif material_dict['type'] == 'log_normal-two_values':
            y_min_log = np.log(float(material_dict['y_min']))
            y_max_log = np.log(float(material_dict['y_max']))
            mean = (y_min_log + y_max_log) / 2.0
            std = (mean - y_min_log) / 2.0

            generator_dict[material] = NormalDistribution(mean=mean,
                                                          std=std,
                                                          log=True,
                                                          )

    final_list = []
    for sample_id in range(n):
        current_case = {}
        for generator in generator_dict:
            current_case[generator] = generator_dict[generator].run()
        final_list.append(current_case)
    if write_to_file:
        df_test = pd.DataFrame(final_list)
        df_test.to_csv('SA_cases.csv', index=False)
    if return_generator:
        return generator_dict
    return final_list





import pathlib

import pandas as pd

from pydelling.preprocessing.mesh_preprocessor import MeshPreprocessor
from pydelling.preprocessing.dfn_preprocessor import DfnPreprocessor
from pydelling.preprocessing.dfn_preprocessor import Fracture
import pydelling.preprocessing.mesh_preprocessor.geometry as geometry
from pydelling.utils.geometry_utils import compute_polygon_area, filter_unique_points
import logging
from tqdm import tqdm
import numpy as np
import dill

logger = logging.getLogger(__name__)


class DfnUpscaler:
    def __init__(self, dfn: DfnPreprocessor,
                 mesh: MeshPreprocessor,
                 parallel=False,
                 save_intersections=False,
                 load_faults:str or pathlib.Path=None,
                 ):
        self.dfn: DfnPreprocessor = dfn
        self.mesh: MeshPreprocessor = mesh
        logger.info('The DFN and mesh objects have been set properly')
        self.all_intersected_points = []
        self.save_intersections = save_intersections
        self.load_faults = load_faults
        self._intersect_dfn_with_mesh(parallel=parallel)

    def _intersect_dfn_with_mesh(self, parallel=False):
        """Runs the DfnUpscaler"""
        logger.info('Upscaling the DFN to the mesh')

        for fracture in tqdm(self.dfn, desc='Intersecting fractures with mesh', total=len(self.dfn)):
            self.find_intersection_points_between_fracture_and_mesh(fracture)
        if not self.load_faults:
            self.find_fault_cells()
        else:
            logger.info(f'Loading fault assignment information from {self.load_faults}')
            with open(self.load_faults, 'rb') as f:
                fault_info = dill.load(f)
                for element in self.mesh.elements:
                    element.associated_faults = fault_info[element.local_id]


        if self.save_intersections:
            import csv
            with open('intersections.csv', 'w') as f:
                writer = csv.writer(f)
                writer.writerow(['x', 'y', 'z'])
                for intersection in self.all_intersected_points:
                    for point in intersection:
                        writer.writerow([point.x, point.y, point.z])

    def find_intersection_points_between_fracture_and_mesh(self, fracture: Fracture, export_stats=False):
        """Finds the intersection points between a fracture and the mesh"""

        intersection_points = []
        kd_tree_filtered_elements = self.mesh.get_closest_mesh_elements(fracture.centroid, distance=fracture.size)
        counter = 0
        elements_filtered = 0
        for element in kd_tree_filtered_elements:
            element: geometry.BaseElement
            counter += 1
            absolute_distance = np.abs(fracture.distance_to_point(element.centroid))
            characteristic_length = np.power(element.volume, 1 / 3)
            if absolute_distance > 1.75 * characteristic_length:
                elements_filtered += 1
                continue

            intersection_points = element.intersect_with_fracture(fracture)
            if self.save_intersections:
                self.all_intersected_points.append(intersection_points)

            intersection_area = compute_polygon_area(intersection_points)
            if intersection_area == None:
                if hasattr(self, 'none_written'):
                    continue
                with open('none_intersections.txt', 'a') as f:
                    for intersection_point in intersection_points:
                        f.write(f'{intersection_point[0]},{intersection_point[1]},{intersection_point[2]}\n')
                self.none_written = True
            elif intersection_area < 0:
                if hasattr(self, 'negative_written'):
                    continue
                with open('negative_intersections.txt', 'a') as f:
                    for intersection_point in intersection_points:
                        f.write(f'{intersection_point[0]},{intersection_point[1]},{intersection_point[2]}\n')
                self.negative_written = True

            fracture.intersection_dictionary[element.local_id] = intersection_area
            if len(intersection_points) > 0:
                element.associated_fractures[fracture.local_id] = {
                    'area': intersection_area,
                    'volume': intersection_area * fracture.aperture,
                    'fracture': fracture,
                }
            n_intersections = len(intersection_points)
            if not n_intersections in self.mesh.find_intersection_stats['intersection_points'].keys():
                self.mesh.find_intersection_stats['intersection_points'][n_intersections] = 0
            self.mesh.find_intersection_stats['intersection_points'][n_intersections] += 1
            self.mesh.find_intersection_stats['total_intersections'] += 1

        self.mesh.is_intersected = True
        return intersection_points

    def find_fault_cells(self, save_fault_cells=True):
        """Finds the fault cells in the mesh"""
        logger.info('Finding fault cells')
        fault_cells = {}
        for fault in self.dfn.faults:

            kd_tree_filtered_elements = self.mesh.get_closest_mesh_elements(fault.centroid, distance=fault.size)

            if len(kd_tree_filtered_elements) == 0:
                continue

            logger.info(f'Processing fault {fault}')
            logger.info(f'Number of elements in the fault: {len(kd_tree_filtered_elements)}')

            kd_tree_centroids = np.array([elem.centroid for elem in kd_tree_filtered_elements])
            fault_plane = Fracture(normal_vector=fault.normal_vector,
                                   x=fault.centroid[0],
                                   y=fault.centroid[1],
                                   z=fault.centroid[2],
                                   aperture=fault.aperture
                                   )
            distance_vec = []
            for kd_centroid in kd_tree_centroids:
                distance_vec.append(fault_plane.distance_to_point(kd_centroid))
            # Filter distances
            distance_vec = np.array(distance_vec)
            distance_vec = np.abs(distance_vec)
            distance_vec = pd.DataFrame(distance_vec)
            distance_vec = distance_vec[distance_vec < fault.aperture * 2].dropna()
            kd_tree_filtered_elements = [kd_tree_filtered_elements[i] for i in distance_vec.index]
            distances = distance_vec.values

            # distances = fault.distance(kd_tree_centroids)
            for element, distance in zip(kd_tree_filtered_elements, distances):
                distance = np.abs(distance)
                if distance < fault.aperture / 2:
                    element.associated_faults[fault.local_id] = {
                        'distance': distance[0],
                    }
                    fault.associated_elements.append(element)

        if save_fault_cells:
            import dill
            for element in self.mesh.elements:
                fault_cells[element.local_id] = element.associated_faults
            with open('fault_cells.pkl', 'wb') as f:
                dill.dump(fault_cells, f)




    def _compute_fracture_volume_in_elements(self):
        # Compute volume of fractures in each element.
        # self.elements.total_fracture_volume = np.zeros([len(elements)])
        for elem in tqdm(self.mesh.elements, desc="Computing fracture volume fractions"):
            for fracture in elem.associated_fractures:
                fracture_dict = elem.associated_fractures[fracture]
                # Attribute of the element: portion of element occupied by fractures.
                elem.total_fracture_volume += fracture_dict['volume']

    def upscale_mesh_porosity(self, matrix_porosity=None):
        # Compute upscaled porosity for each element.
        self._compute_fracture_volume_in_elements()
        upscaled_porosity = {}
        for elem in tqdm(self.mesh.elements, desc="Upscaling porosity"):
            upscaled_porosity[elem.local_id] = (elem.total_fracture_volume / elem.volume)  # + (
            # matrix_porosity[elem] * (1 - (elem.total_fracture_volume / elem.volume)))
            # if elem.total_fracture_volume > 0:
            #     pass
            #     #print(elem.total_fracture_volume, elem.volume, matrix_porosity[elem])


        for fault in self.dfn.faults:
            for element in fault.associated_elements:
                upscaled_porosity[elem.local_id] = fault.porosity

        #POROSITY POST-PROCESSING
        for elem in tqdm(self.mesh.elements, desc="Post processing upscaled porosity"):
            upscaled_porosity[elem.local_id] = upscaled_porosity[elem.local_id] * (1.53) * 1 / 0.3842

        vtk_porosity = np.asarray(self.mesh.elements)
        for local_id in upscaled_porosity:
            vtk_porosity[local_id] = upscaled_porosity[local_id]

        self.mesh.cell_data['upscaled_porosity'] = [vtk_porosity.tolist()]
        self.upscaled_porosity = upscaled_porosity

        return upscaled_porosity

    def upscale_mesh_storativity(self, matrix_storativity=None):

        upscaled_storativity = {}

        for elem in tqdm(self.mesh.elements, desc="Upscaling fractures storativity"):
            upscaled_storativity[elem.local_id] = 0.0
            for frac_name in elem.associated_fractures:
                frac_dict = elem.associated_fractures[frac_name]
                frac = frac_dict['fracture']
                upscaled_storativity[elem.local_id] += frac.storativity

        for fault in self.dfn.faults:
            for element in fault.associated_elements:
                upscaled_storativity[elem.local_id] = fault.storativity


        #STORATIVITY POST-PROCESSING
        for elem in tqdm(self.mesh.elements, desc="Post processing upscaled porosity"):
            upscaled_storativity[elem.local_id] = upscaled_storativity[elem.local_id] * (1.53) * 0.998 / 0.187

        vtk_storativity = np.asarray(self.mesh.elements)
        for local_id in upscaled_storativity:
            vtk_storativity[local_id] = upscaled_storativity[local_id]

        self.mesh.cell_data['upscaled_storativity'] = [vtk_storativity.tolist()]
        self.upscaled_storativity = upscaled_storativity

        return upscaled_storativity

    def export_fault_distances(self):
        """Exports the fault distances"""
        distance = {}

        for elem in tqdm(self.mesh.elements, desc="Computing distances"):
            distance[elem.local_id] = 0
            for fault in elem.associated_faults:
                fault_dict = elem.associated_faults[fault]
                distance[elem.local_id] += fault_dict['distance'] if 'distance' in fault_dict else 0

        vtk_distance = np.asarray(self.mesh.elements)
        for local_id in distance:
            vtk_distance[local_id] = distance[local_id]

        self.mesh.cell_data['distance'] = [vtk_distance.tolist()]

        self.distance = distance

    def upscale_mesh_permeability(self, matrix_permeability=None, rho=1000, g=9.8, mu=8.9e-4,
                                  mode='full_tensor'):

        matrix_permeability = {}

        for elem in tqdm(self.mesh.elements, desc="Creating permeability tensor for dummy anisotropic case"):
            matrix_permeability[elem.local_id] = np.ones([3, 3]) * 0.0

        matrix_permeability_tensor = matrix_permeability

        # Check correct size of matrix_permeability.
        # matrix_permeability_tensor = np.zeros(len(self.elements))
        #
        # if len(matrix_permeability) != len(self.elements):
        #     print("Incorrect size for matrix permeability. Size of variable doesn't match number of elements in the mesh.")
        #     break
        # else:
        #     for elem in tqdm(self.elements, desc="Check size of matrix permeability input"):
        #         if len(matrix_permeability[elem]) == 3:
        #             if np.shape(matrix_permeability[elem]) == (3,3):
        #                 print("Matrix Permeability Tensor (3,3) for Anisotropic case.")
        #                 continue
        #             else:
        #                 print("Matrix Permeability Tensor must be an np.array([3,3]) for Anisotropic case.")
        #         elif len(matrix_permeability[elem]) == 1:
        #             print("Matrix Permeability for Isotropic case.")
        #             matrix_permeability_tensor[elem] = np.zeros([3,3])
        #             matrix_permeability_tensor[elem][0,0] = matrix_permeability[elem]
        #             continue
        #         else:
        #             print("Incorrect Matrix Permeability Tensor. Must be an np.array([3,3]) for use in Anisotropic case or a single float/int for use in Isotropic case.")
        #             continue

        # UPSCALED PERMEABILITY
        fracture_hk = {}
        fault_hk = {}
        upscaled_hk = {}

        # For each fracture, compute permeability tensor,
        # and add it to the elements intersected by the fracture.
        for elem in tqdm(self.mesh.elements, desc="Upscaling fractures permeability"):
            fracture_hk[elem.local_id] = np.zeros([3, 3])
            upscaled_hk[elem.local_id] = np.zeros([3, 3])
            fault_hk[elem.local_id] = np.zeros([3, 3])

            for frac_name in elem.associated_fractures:
                frac_dict = elem.associated_fractures[frac_name]
                frac = frac_dict['fracture']
                # n1 = math.cos(frac.dip * (math.pi / 180)) * math.sin(frac.dip_dir * (math.pi / 180))
                # n2 = math.cos(frac.dip * (math.pi / 180)) * math.cos(frac.dip_dir * (math.pi / 180))
                # n3 = -1 * math.sin(frac.dip * (math.pi / 180))
                n1 = frac.unit_normal_vector[0]
                n2 = frac.unit_normal_vector[1]
                n3 = frac.unit_normal_vector[2]
                # frac.hk = ((frac.aperture ** 2) * rho * g) / (12 * mu)
                frac.hk = frac.transmissivity / frac.aperture

                if 'mode' == 'isotropy':
                    # Add fracture permeability, weighted by the area that the fracture occupies in the element.
                    fracture_hk[elem.local_id][0, 0] += frac.hk * (frac_dict['volume'] / elem.volume)  # Kxx

                else:  # 'anisotropy' in 'mode':
                    perm_tensor = np.zeros([3, 3])
                    # for i in range(1, 4):
                    #    for j in range(1, 4):
                    # Compute tensor
                    perm_tensor[0, 0] = frac.hk * ((n2 ** 2) + (n3 ** 2))
                    perm_tensor[0, 1] = frac.hk * (-1) * n1 * n2
                    perm_tensor[0, 2] = frac.hk * (-1) * n1 * n3
                    perm_tensor[1, 1] = frac.hk * ((n3 ** 2) + (n1 ** 2))
                    perm_tensor[1, 2] = frac.hk * (-1) * n2 * n3
                    perm_tensor[2, 2] = frac.hk * ((n1 ** 2) + (n2 ** 2))

                    if 'mode' == 'anisotropy_principals':
                        eigen_perm_tensor = np.diag(np.linalg.eig(perm_tensor)[0])
                        perm_tensor = eigen_perm_tensor

                    # Add fracture permeability, weighted by the area that the fracture occupies in the element.
                    fracture_hk[elem.local_id][0, 0] += (perm_tensor[0, 0] * frac_dict['volume'] / elem.volume)
                    fracture_hk[elem.local_id][0, 1] += (perm_tensor[0, 1] * frac_dict['volume'] / elem.volume)
                    fracture_hk[elem.local_id][0, 2] += (perm_tensor[0, 1] * frac_dict['volume'] / elem.volume)
                    fracture_hk[elem.local_id][1, 0] += (perm_tensor[0, 1] * frac_dict['volume'] / elem.volume)
                    fracture_hk[elem.local_id][1, 1] += (perm_tensor[1, 1] * frac_dict['volume'] / elem.volume)
                    fracture_hk[elem.local_id][1, 2] += (perm_tensor[1, 2] * frac_dict['volume'] / elem.volume)
                    fracture_hk[elem.local_id][2, 0] += (perm_tensor[0, 1] * frac_dict['volume'] / elem.volume)
                    fracture_hk[elem.local_id][2, 1] += (perm_tensor[1, 2] * frac_dict['volume'] / elem.volume)
                    fracture_hk[elem.local_id][2, 2] += (perm_tensor[2, 2] * frac_dict['volume'] / elem.volume)

            # Sum permeability contribution from fractures and from matrix.
            upscaled_hk[elem.local_id] = fracture_hk[elem.local_id] + matrix_permeability_tensor[elem.local_id] * (
                    1 - (elem.total_fracture_volume / elem.volume))

            if len(elem.associated_faults) > 0:

                for fault_name in elem.associated_faults:
                    fault_dict = elem.associated_faults[fault_name]
                    fault = fault_dict['fault']
                    # n1 = math.cos(frac.dip * (math.pi / 180)) * math.sin(frac.dip_dir * (math.pi / 180))
                    # n2 = math.cos(frac.dip * (math.pi / 180)) * math.cos(frac.dip_dir * (math.pi / 180))
                    # n3 = -1 * math.sin(frac.dip * (math.pi / 180))
                    n1 = fault.unit_normal_vector[0]
                    n2 = fault.unit_normal_vector[1]
                    n3 = fault.unit_normal_vector[2]
                    # frac.hk = ((frac.aperture ** 2) * rho * g) / (12 * mu)
                    fault.hk = fault.transmissivity / fault.aperture

                if 'mode' == 'isotropy':
                    # Add fault permeability. Element porosity equals to 1 when intersected by faults.
                    fault_hk[elem.local_id][0, 0] += fault.hk

                else:  # 'anisotropy' in 'mode':
                    perm_tensor = np.zeros([3, 3])
                    # for i in range(1, 4):
                    #    for j in range(1, 4):
                    # Compute tensor
                    perm_tensor[0, 0] = fault.hk * ((n2 ** 2) + (n3 ** 2))
                    perm_tensor[0, 1] = fault.hk * (-1) * n1 * n2
                    perm_tensor[0, 2] = fault.hk * (-1) * n1 * n3
                    perm_tensor[1, 1] = fault.hk * ((n3 ** 2) + (n1 ** 2))
                    perm_tensor[1, 2] = fault.hk * (-1) * n2 * n3
                    perm_tensor[2, 2] = fault.hk * ((n1 ** 2) + (n2 ** 2))

                    if 'mode' == 'anisotropy_principals':
                        eigen_perm_tensor = np.diag(np.linalg.eig(perm_tensor)[0])
                        perm_tensor = eigen_perm_tensor

                    # Add fault permeability. Element porosity equals to 1 when intersected by faults.
                    fault_hk[elem.local_id][0, 0] += perm_tensor[0, 0]
                    fault_hk[elem.local_id][0, 1] += perm_tensor[0, 1]
                    fault_hk[elem.local_id][0, 2] += perm_tensor[0, 1]
                    fault_hk[elem.local_id][1, 0] += perm_tensor[0, 1]
                    fault_hk[elem.local_id][1, 1] += perm_tensor[1, 1]
                    fault_hk[elem.local_id][1, 2] += perm_tensor[1, 2]
                    fault_hk[elem.local_id][2, 0] += perm_tensor[0, 1]
                    fault_hk[elem.local_id][2, 1] += perm_tensor[1, 2]
                    fault_hk[elem.local_id][2, 2] += perm_tensor[2, 2]

                # Sum permeability contribution from faults.
                upscaled_hk[elem.local_id] = upscaled_hk[elem.local_id] + fault_hk[elem.local_id]

        # Export values to VTK
        vtk_kxx = np.asarray(self.mesh.elements)
        vtk_kyy = np.asarray(self.mesh.elements)
        vtk_kzz = np.asarray(self.mesh.elements)
        vtk_kxy = np.asarray(self.mesh.elements)
        vtk_kxz = np.asarray(self.mesh.elements)
        vtk_kyz = np.asarray(self.mesh.elements)
        for local_id in upscaled_hk:
            vtk_kxx[local_id] = upscaled_hk[local_id][0, 0]
            vtk_kyy[local_id] = upscaled_hk[local_id][1, 1]
            vtk_kzz[local_id] = upscaled_hk[local_id][2, 2]
            vtk_kxy[local_id] = upscaled_hk[local_id][0, 1]
            vtk_kxz[local_id] = upscaled_hk[local_id][0, 2]
            vtk_kyz[local_id] = upscaled_hk[local_id][1, 2]

        self.mesh.cell_data['Kxx'] = [vtk_kxx.tolist()]
        self.mesh.cell_data['Kyy'] = [vtk_kyy.tolist()]
        self.mesh.cell_data['Kzz'] = [vtk_kzz.tolist()]
        self.mesh.cell_data['Kxy'] = [vtk_kxy.tolist()]
        self.mesh.cell_data['Kxz'] = [vtk_kxz.tolist()]
        self.mesh.cell_data['Kyz'] = [vtk_kyz.tolist()]

        self.upscaled_permeability = upscaled_hk

        return upscaled_hk

    def to_vtk(self, filename):
        """Exports the mesh and the upscaled variables to VTK"""
        self.mesh.to_vtk(filename)

    def porosity_to_csv(self, filename='./porosity.csv'):
        """Exports porosity values to csv

        Returns:

        """
        import csv
        logger.info(f"Exporting porosity to {filename}")
        with open(filename, 'w') as csvfile:
            writer = csv.writer(csvfile)
            for element_id in self.upscaled_porosity:
                centroid = self.mesh.elements[element_id].centroid
                porosity = self.upscaled_porosity[element_id]
                writer.writerow([centroid[0], centroid[1], centroid[2], porosity])

    def plot_porosity_histogram(self, filename='upscaled_porosity_histogram.png'):
        """Plots the upscaled porosity histogram."""
        logger.info(f'Plotting upscaled porosity histogram to {filename}')
        import matplotlib.pyplot as plt
        fig, ax = plt.subplots()
        plt.hist([self.upscaled_porosity[element_id] for element_id in self.upscaled_porosity], bins=100)
        return fig, ax

    def plot_hkx_histogram(self, filename='upscaled_hkx_histogram.png'):
        """Plots the upscaled hkx histogram."""
        logger.info(f'Plotting upscaled hkx histogram to {filename}')
        import matplotlib.pyplot as plt
        fig, ax = plt.subplots()
        plt.hist([self.upscaled_permeability[element_id][0, 0] for element_id in self.upscaled_permeability], bins=100)
        return fig, ax

    def plot_storativity_histogram(self, filename='upscaled_storativity_histogram.png'):
        """Plots the upscaled storativity histogram."""
        logger.info(f'Plotting upscaled storativity histogram to {filename}')
        import matplotlib.pyplot as plt
        fig, ax = plt.subplots()
        plt.hist([self.upscaled_storativity[element_id] for element_id in self.upscaled_storativity], bins=100)
        return fig, ax

    def export_intersection_stats(self, filename='intersection_stats.txt'):
        # Export the run_stats dictionary to file
        assert self.mesh.is_intersected, 'The mesh has not been intersected yet.'
        import json
        with open('run_stats.json', 'w') as fp:
            json.dump(self.mesh.find_intersection_stats, fp)

    def save(self, filename):
        """Save a copy of the class on a serialized pickle object"""
        import pickle
        with open(filename, 'wb') as f:
            pickle.dump(self, f)

    @classmethod
    def load(cls, filename):
        """Load a serialized pickle object"""
        import pickle
        with open(filename, 'rb') as f:
            return pickle.load(f)





from pydelling.paraview_processor.filters import BaseFilter
try:
    from paraview.simple import *
except:
    pass
import pandas as pd
from vtk.util.numpy_support import vtk_to_numpy
from functools import wraps


def convert_to_table(func):
    @wraps(func)
    def wrapper():

        func()
    return wrapper


class CsvReaderFilter(BaseFilter):
    filter_type: str = "VTK_reader"
    counter: int = 0
    x_min: float
    x_max: float
    y_min: float
    y_max: float
    z_min: float
    z_max: float

    def __init__(self, filename, name, coordinate_labels=("x", "y", "z")):
        super().__init__(name=name)
        self.filter = CSVReader(FileName=str(filename))
        CsvReaderFilter.counter += 1
        self.coordinate_labels = coordinate_labels
        self.filter = self.convert_table_to_points()
        self.set_ranges()

    def set_ranges(self):
        self.x_min = self.mesh_points.min()["x"]
        self.x_max = self.mesh_points.max()["x"]
        self.y_min = self.mesh_points.min()["y"]
        self.y_max = self.mesh_points.max()["y"]
        self.z_min = self.mesh_points.min()["z"]
        self.z_max = self.mesh_points.max()["z"]

    def convert_table_to_points(self):
        _table_to_points = TableToPoints(Input=self.filter)
        _table_to_points.XColumn = self.coordinate_labels[0]
        _table_to_points.YColumn = self.coordinate_labels[1]
        _table_to_points.ZColumn = self.coordinate_labels[2]
        return _table_to_points


import os
import h5py
import numpy as np


class BaseWriter:
    info: dict
    def __init__(self, filename=None, var_name=None, data=None, region_name=None, **kwargs):
        self.data_loaded = False
        self.a_min = None
        self.a_max = None
        self.__dict__.update(**kwargs)
        self.filename = filename if filename else "test.dat"
        if var_name:
            self.var_name = var_name
        if region_name:
            self.region_name = region_name
        if data is not None:
            self.data = data
            self.data_loaded = True

    def set_data_limits(self, a_min=None, a_max=None):
        if a_min == 'None':
            a_min = None
        if a_min is not None:
            self.a_min = float(a_min)
        if a_max == 'None':
            a_max = None
        if a_max is not None:
            self.a_max = float(a_max)

    def apply_data_limits(self):
        print(f"Applying data limits to {self.data} with minimum value: {self.a_min} and maximum value: {self.a_max}")
        if not (self.a_min is None and self.a_max is None):
            try:
                self.data = np.clip(self.data, a_min=self.a_min, a_max=self.a_max)
                print(f"Resulting array: {self.data} Min value: {np.min(self.data)} Max value: {np.max(self.data)}")
            except Exception as e:
                print(f"INFO: Cannot apply minimum and maximum values")
                print(f"ERROR: {e}")

    def load_data(self, var_name=None, data=None, apply_data_limits=True):
        """
        Load data to export
        :return:
        """
        self.wipe_data()
        if var_name is not None and data is not None:
            self.var_name = var_name
            self.data = data
            self.data_loaded = True
        if self.has_data_loaded() and apply_data_limits:
            self.apply_data_limits()
        return self.data_loaded

    def wipe_data(self):
        self.var_name = None
        self.data_loaded = False
        self.data = None

    def check_data(self):
        return self.data_loaded

    def run(self, filename=None):
        if filename is not None:
            self.filename = filename
        if self.check_data():
            if not os.path.exists(self.filename):
                temp_writer = open(self.filename, "w")
                temp_writer.close()
            with open(self.filename) as temp_writer:
                if type(self.data) == np.ndarray:
                    np.savetxt(self.filename, self.data)
            self.info["writer"] = {"filename": self.filename}
        else:
            print("Couldn't find data to dump!")

    def remove_output_file(self, filename=None):
        if filename is None:
            filename = self.filename
        else:
            self.filename = filename
        try:
            os.remove(filename)
        except FileNotFoundError as ef:
            print(f"Nothing to delete!")
        except PermissionError as perr:
            print(f"ERROR: Cannot delete file: {perr}")
            exit(1)

    def has_data_loaded(self):
        return self.data_loaded




import unittest
from pydelling.readers import iGPReader
from pydelling.config import config
from pydelling.utils import test_data_path
import numpy as np


class iGPReaderCase(unittest.TestCase):
    """
    This class tests the iGPReader[pydelling.readers.iGPReader.io.iGPReader.iGPReader] class implementation
    """
    def setUp(self) -> None:
        self.igp_reader = iGPReader(test_data_path() / "test_implicit_to_explicit.gid",
                                            project_name="test-implicit_to_explicit"
                                            )

    def test_implicit_to_explicit(self):
        self.igp_reader.build_mesh_data()
        self.igp_reader.implicit_to_explicit()

    def test_limits(self):
        min_x = self.igp_reader.min_x
        min_nodes_x = self.igp_reader.coords_min_x
        explicit_min_x = np.min(self.igp_reader.centroids[:, 0])
        explicit_min_x_nodes = np.min(self.igp_reader.nodes[:, 0])

        self.assertEqual(min_x, explicit_min_x)
        self.assertEqual(min_nodes_x, explicit_min_x_nodes)

        max_x = self.igp_reader.max_x
        max_nodes_x = self.igp_reader.coords_max_x
        explicit_max_x = np.max(self.igp_reader.centroids[:, 0])
        explicit_max_x_nodes = np.max(self.igp_reader.nodes[:, 0])

        self.assertEqual(max_x, explicit_max_x)
        self.assertEqual(max_nodes_x, explicit_max_x_nodes)

        min_y = self.igp_reader.min_y
        min_nodes_y = self.igp_reader.coords_min_y
        explicit_min_y = np.min(self.igp_reader.centroids[:, 1])
        explicit_min_y_nodes = np.min(self.igp_reader.nodes[:, 1])

        self.assertEqual(min_y, explicit_min_y)
        self.assertEqual(min_nodes_y, explicit_min_y_nodes)

        max_y = self.igp_reader.max_y
        max_nodes_y = self.igp_reader.coords_max_y
        explicit_max_y = np.max(self.igp_reader.centroids[:, 1])
        explicit_max_y_nodes = np.max(self.igp_reader.nodes[:, 1])

        self.assertEqual(max_y, explicit_max_y)
        self.assertEqual(max_nodes_y, explicit_max_y_nodes)

        min_z = self.igp_reader.min_z
        min_nodes_z = self.igp_reader.coords_min_z
        explicit_min_z = np.min(self.igp_reader.centroids[:, 2])
        explicit_min_z_nodes = np.min(self.igp_reader.nodes[:, 2])

        self.assertEqual(min_z, explicit_min_z)
        self.assertEqual(min_nodes_z, explicit_min_z_nodes)

        max_z = self.igp_reader.max_z
        max_nodes_z = self.igp_reader.coords_max_z
        explicit_max_z = np.max(self.igp_reader.centroids[:, 2])
        explicit_max_z_nodes = np.max(self.igp_reader.nodes[:, 2])

        self.assertEqual(max_z, explicit_max_z)
        self.assertEqual(max_nodes_z, explicit_max_z_nodes)


if __name__ == '__main__':
    unittest.main()



"""
Centroid file reader
"""
import numpy as np
from .BaseReader import BaseReader
import logging
import pandas as pd

logger = logging.getLogger(__name__)


class CentroidReader(BaseReader):
    """This class reads a data file described by a set of centroids"""
    def __init__(self, filename, var_pos=3, var_name="var", var_type=np.float32, centroid_pos=(0, 3), header=False, separator=None):
        self.var_pos = var_pos
        self.var = None
        self.var_name = var_name
        self.var_type = var_type
        self.centroid_pos = centroid_pos
        self.header = header
        self.split_key = separator
        if self.var_pos:
            logger.info(f"Reading data using centroids in positions [{self.centroid_pos[0]}, {self.centroid_pos[1] - 1}] and data in position {self.var_pos}")
        else:
            logger.info(f"Reading data using centroids in positions [{self.centroid_pos[0]}, {self.centroid_pos[1] - 1}]")

        super().__init__(filename, var_pos=var_pos,
                         var_name=var_name,
                         var_type=var_type,
                         centroid_pos=centroid_pos,
                         header=header)

    def read_file(self, opened_file):
        """
        Reads the data and stores it inside the class
        """
        logger.info(f"Reading centroid file from {self.filename}")
        temp_centroid = []
        temp_id = []
        for line in opened_file.readlines():
            if self.split_key:
                data_row = line.split(self.split_key)
            else:
                data_row = line.split()
            temp_centroid.append(data_row[self.centroid_pos[0]:self.centroid_pos[1] + 1])
            if self.var_pos:
                temp_id.append([data_row[self.var_pos]])
        self.data = np.array(temp_centroid, dtype=np.float32)
        if self.var_pos:
            self.var = np.array(temp_id, dtype=self.var_type)


    def read_header(self):
        """
        TODO: Add the header reader of the centroid file
        Reads the header of the file
        :return:
        """
        pass

    def get_data(self, as_dataframe=False) -> pd.DataFrame:
        """
        Outputs the data
        :return: np.ndarray object containing centroid information and variable output
        """
        if as_dataframe:
            if self.var_pos:
                return pd.DataFrame(self.data, columns=['x', 'y', 'z', f"{self.var_name}"])
            else:
                return pd.DataFrame(self.data, columns=['x', 'y', 'z'])
        else:
            return self.data


    def build_info(self):
        """
        Generates a dictionary containing the basic info of the read data
        :return:
        """
        self.info["reader"] = {"n_cells": self.data.shape[0],
                     "filename": self.filename,
                     "var_name": self.var_name,
                     "var_position": self.var_pos}

    def to_csv(self, output_file, delimiter=","):
        """
        Writes the data into a csv file
        :param output_file:
        :return:
        """
        logger.info(f"Starting dump into {output_file}")
        np.savetxt(output_file, self.get_data(), delimiter=delimiter)
        logger.info(f"The data has been properly exported to the {output_file} file")

    def shift(self, direction=np.array([0.0, 0.0, 0.0])):
        """Shifts the data coordinates on the given direction
        Args:
            direction: direction vector
        """
        if type(direction) is list:
            direction = np.array(direction)
        self.data[:, 0:3] += direction




import unittest
from pydelling.readers import SmeshReader
from pydelling.utils import test_data_path


class TestSMeshMeshReader(unittest.TestCase):
    def test_read_data(self):
        smesh_reader = SmeshReader(test_data_path() / 'smesh_test_data.smesh')

        self.assertEqual(smesh_reader.n_nodes, 4)
        self.assertEqual(smesh_reader.n_elements, 1)




if __name__ == '__main__':
    unittest.main()


from pydelling.config import config


class BaseElement:
    def __init__(self, node_ids, node_coords, element_type_n, local_id, centroid_coords=None):
        self.nodes = node_ids  # Node id set
        self.coords = node_coords  # Coordinates of each node
        self.centroid_coords = centroid_coords
        self.type = config.globals.element_dict[element_type_n]  # Element type
        self.n_type = element_type_n  # Number of node_ids for an element
        self.local_id = local_id  # Element id
        self.faces = {}  # Dictionary to store face information

    def __repr__(self):
        print("### Element info ###")
        print(f"Element ID: {self.local_id}")
        print(f"Number of nodes: {self.n_type}")
        print(f"Element type: {self.type}")
        print(f"Node list: {self.nodes}")
        print("### End element info ###")

        print("### Face info ###")
        for face in self.faces:
            print(f"{face}: {self.faces[face].coords}")
        print("### End face info ###")

    def print_element_info(self):
        print("### Element info ###")
        print(f"Element ID: {self.local_id}")
        print(f"Number of nodes: {self.n_type}")
        print(f"Element type: {self.type}")
        print(f"Node list: {self.nodes}")
        print("### End element info ###")

    def print_face_info(self):
        print("### Face info ###")
        for face in self.faces:
            print(f"{face}: {self.faces[face].coords}")
        print("### End face info ###")



from __future__ import annotations
import numpy as np
from . import Scalar
from .BasePrimitive import BasePrimitive
from typing import *


class Point(np.ndarray):
    def __new__(cls, input_array):
        # Input array is an already formed ndarray instance
        # We first cast to be our class type
        if len(input_array) == 1:
            raise ValueError("Point must have 2 or 3 coordinates")
        elif len(input_array) == 2:
            obj = np.asarray(input_array).view(cls)
        elif len(input_array) == 3:
            obj = np.asarray(input_array).view(cls)
        else:
            raise ValueError("Point must have 2 or 3 coordinates")
        return obj

    def distance(self, p: Point):
        """Computes euclidean distance between two points"""
        diff = self - p
        return float(np.sqrt((diff ** 2).sum()))

    def __repr__(self):
        return f"Point({self})"


    @property
    def x(self):
        return self[0]

    @property
    def y(self):
        return self[1]

    @property
    def z(self):
        return self[2]




import numpy as np

from pydelling.readers.iGPReader.io import BaseReader
from pydelling.readers.iGPReader.utils import get_output_path
from pydelling.config import config
import h5py
import logging

logger = logging.getLogger(__name__)


class AscReader(BaseReader):
    """
     Class that contains functions to read a rasterized file in .asc format
     """

    def __init__(self, filename):
        self.filename = filename
        self.info_dict = {}
        self.opened_file = open(self.filename, "r")
        self.xydata_computed = False
        self.read_header()
        self.build_data_structure()
        self.read_data()
        logger.debug(f"CSV data has been read from {self.filename}")
        logger.debug(f"{self.info_dict}")

    def read_header(self, n_header=6):
        for i in range(0, n_header):
            line = self.opened_file.readline().split()
            self.info_dict[line[0]] = float(line[1])

    def read_data(self):
        for id, line in enumerate(self.opened_file.readlines()):
            self.data[id] = line.split()

    def build_data_structure(self):
        """
        Function that creates the internal data structure of the raster file
        """
        assert self.info_dict is not {}
        self.data = np.zeros(shape=(int(self.info_dict["nrows"]), int(self.info_dict["ncols"])))
        x_range = np.arange(self.info_dict["xllcorner"], self.info_dict["nrows"] * self.info_dict["cellsize"],
                            self.info_dict["cellsize"])
        y_range = np.arange(self.info_dict["yllcorner"], self.info_dict["ncols"] * self.info_dict["cellsize"],
                            self.info_dict["cellsize"])
        self.x_mesh, self.y_mesh = np.meshgrid(x_range, y_range)
        self.y_mesh = np.flipud(self.y_mesh)  # To fit into the .asc format criteria

    def rebuild_x_y(self):
        x_range = np.arange(self.info_dict["xllcorner"], self.info_dict["nrows"] * self.info_dict["cellsize"],
                            self.info_dict["cellsize"])
        y_range = np.arange(self.info_dict["yllcorner"], self.info_dict["ncols"] * self.info_dict["cellsize"],
                            self.info_dict["cellsize"])
        self.x_mesh, self.y_mesh = np.meshgrid(x_range, y_range)

    def dump_to_xydata(self):
        ndata = int(self.info_dict["nrows"] * self.info_dict["ncols"])
        self.xydata = np.zeros(shape=(ndata, 3))
        x_mesh_flatten = self.x_mesh.flatten()
        y_mesh_flatten = self.y_mesh.flatten()
        for id, data in enumerate(self.data.flatten()):
            self.xydata[id] = (x_mesh_flatten[id], y_mesh_flatten[id], data)
        self.xydata_computed = True
        return self.xydata

    def dump_to_csv(self, output_file):
        """
        Function that writes the ratser data into a csv file
        :param output_file:
        :return:
        """
        logger.info(f"Writing into {output_file}")
        if not self.xydata_computed:
            xydata = self.dump_to_xydata()
        else:
            xydata = self.xydata
        f = open(get_output_path() / output_file, "w")
        for data in xydata:
            if float(data[2]) == -9999.0:
                continue
            f.write(f"{data[0]},{data[1]},{data[2]}\n")
        f.close()

    def dump_to_wsv(self, output_file):
        logger.info(f"Writing into {output_file}")
        if not self.xydata_computed:
            xydata = self.dump_to_xydata()
        else:
            xydata = self.xydata
        f = open(output_file, "w")
        for data in xydata:
            f.write(f"{data[0]} {data[1]} {data[2]}\n")
        f.close()

    def dump_to_asc(self, output_file):
        logger.info(f"Writing into {output_file}")
        file = open(output_file, "w")
        self.write_asc_header(file)
        self.write_asc_data(file)
        file.close()

    def write_asc_header(self, file):
        # assert isinstance(file, type(open)), "is not a correct file"
        for head in self.info_dict:
            file.write(f"{head} {self.info_dict[head]}\n")

    def write_asc_data(self, file):
        np.savetxt(file, self.data, fmt="%3.2f")

    def export_to_gridded_dataset(self, filename=None, attrs=None):
        """This method exports the asc data to the PFLOTRAN gridded dataset hdf5 file format"""
        filename = get_output_path() / filename if filename else get_output_path() / "gridded_dataset.h5"
        growth_factor = config.general.raster_growth_factor if config.general.raster_growth_factor else 1.0
        if config.general.raster_growth_factor:
            logger.warning(f'Scaling the dx and dy values of the gridded dataset by {config.general.raster_growth_factor}')
        with h5py.File(filename, "w") as hdf5_file:
            hdf5_group = hdf5_file.create_group(filename.stem)
            hdf5_dataset = hdf5_group.create_dataset("Data", data=self.data)
            hdf5_group.attrs.create("Dimension", ["XY"], dtype="S3")
            hdf5_group.attrs["Discretization"] = [self.info_dict["cellsize"]*growth_factor, self.info_dict["cellsize"]*growth_factor]
            hdf5_group.attrs["Origin"] = [self.info_dict["xllcorner"], self.info_dict["yllcorner"]]
            hdf5_group.attrs["Cell Centered"] = [True]
            # hdf5_group.attrs["Interpolation Method"] = "STEP"
            # hdf5_group.attrs["Max Buffer Size"] = 1
        logger.info(f"Gridded dataset has been exported at {filename}")

    def change_data(self, data: np.array):
        self.data = data

    def downsample_data(self, slice_factor=2):
        '''
        This module downsamples the data based on a constant stride in each direction
        :param slice_factor: factor to stride the matrix in each dimension

        :return: downsampled dataset
        '''
        self.data = self.data[0::slice_factor, 0::slice_factor]
        self.info_dict["nrows"] = self.data.shape[0]
        self.info_dict["ncols"] = self.data.shape[1]
        self.info_dict["cellsize"] *= slice_factor
        AscReader.rebuild_x_y(self)
        logger.info("Data has been downsampled, these are the new settings:")
        logger.info(f"{self.info_dict}")



from __future__ import annotations

import numpy as np

from .BasePrimitive import BasePrimitive
from .Point import Point
from typing import List


class Line(BasePrimitive):
    def __init__(self, p1: np.ndarray or Point or List = None, p2: np.ndarray or Point or List = None,
                 direction_vector: np.ndarray or List = None):
        if p1 is not None and p2 is not None:
            p1 = Point(p1)
            p2 = Point(p2)
            self.direction_vector = (p2 - p1) / np.linalg.norm(p2 - p1)
            if isinstance(p1, Point):
                self.p = p1
            else:
                self.p = Point(p1)

        elif direction_vector is not None and (p1 is not None or p2 is not None):
            self.direction_vector = direction_vector / np.linalg.norm(direction_vector)

            if p1 is not None:
                if isinstance(p1, Point):
                    self.p = p1
                else:
                    self.p = Point(p1)
            elif p2 is not None:
                if isinstance(p2, Point):
                    self.p = p2
                else:
                    self.p = Point(p2)

    def is_parallel(self, line: Line):
        if np.isclose(np.dot(self.direction_vector, line.direction_vector), 1):
            return True
        if np.isclose(np.dot(self.direction_vector, line.direction_vector), -1):
            return True

    def angle(self, line: Line):
        return np.arccos(np.dot(self.direction_vector, line.direction_vector) /
                         (np.linalg.norm(self.direction_vector) * np.linalg.norm(line.direction_vector)))

    def intersect(self, primitive: BasePrimitive):
        from .intersections import intersect_line_line, intersect_plane_line

        if primitive.__class__.__name__ == "Line":
            return intersect_line_line(line_1=self, line_2=primitive)
        elif primitive.__class__.__name__ == "Plane":
            return intersect_plane_line(line=self, plane=primitive)



        else:
            raise NotImplementedError(f"Intersection with {type(primitive)} is not implemented")


    def __repr__(self):
        return f"Line(point: {self.p}, direction_vector: {self.direction_vector})"

    def __str__(self):
        return f"Line(point: {self.p}, direction_vector: {self.direction_vector})"



import numpy as np

from . import Line, Point, BasePrimitive


class Segment(Line):
    def __init__(self, p1, p2):
        super().__init__(p1, p2)
        self.p1 = Point(p1)
        self.p2 = Point(p2)
        self.displacement = self.p2 - self.p1

    def __repr__(self):
        return f"Segment(p1: {self.p1}, p2: {self.p2})"

    def __str__(self):
        return f"Segment(p1: {self.p1}, p2: {self.p2})"

    @property
    def length(self):
        return np.sqrt(np.sum(self.displacement ** 2))


    def intersect(self, primitive: BasePrimitive):
        from .intersections import intersect_plane_segment

        if primitive.__class__.__name__ == "Plane":
            return intersect_plane_segment(segment=self, plane=primitive)

        else:
            raise NotImplementedError(f"Intersection of {type(self)} with {type(primitive)} is not implemented")

    def contains(self, point: Point, rel=0.05):
        value = self.p1.distance(point) + self.p2.distance(point) / self.length
        if abs(1 - value) < rel:
            return True
        else:
            return False





import numpy as np
from .BaseReader import BaseReader
import logging

logger = logging.getLogger(__name__)


class RasterFileReader(BaseReader):
    """
    Class that contains functions to read a rasterized file in .asc format
    """

    def open_file(self, filename, n_header=6):
        with open(filename) as opened_file:
            if self.header:
                opened_file.readline()  # For now, skips the header if it has
            self.read_file(opened_file, n_header)
        self.build_info()

    def read_file(self, opened_file, n_header=6):
        self.read_header(opened_file, n_header=n_header)
        self.build_structure()
        self.read_data(opened_file)
        logger.info(f"Reading ASC raster file from {opened_file}")

    def read_header(self, opened_file, n_header=6):
        for i in range(0, n_header):
            line = opened_file.readline().split()
            self.info["reader"][line[0]] = float(line[1])

    def read_data(self, opened_file):
        for id, line in enumerate(opened_file.readlines()):
            self.data[id] = np.array(line.split(), dtype=np.float32)

    def build_structure(self):
        assert self.info is not {}
        self.reader_info = self.info["reader"]
        self.xydata_computed = False
        self.data = np.zeros(shape=(int(self.reader_info["nrows"]), int(self.reader_info["ncols"])))

        if 'cellsize' in self.reader_info:
            x_range = np.arange(self.reader_info["xllcorner"], self.reader_info["xllcorner"] + self.reader_info["nrows"] * self.reader_info["cellsize"],
                                self.reader_info["cellsize"])
            y_range = np.arange(self.reader_info["yllcorner"], self.reader_info["yllcorner"] + self.reader_info["ncols"] * self.reader_info["cellsize"],
                                self.reader_info["cellsize"])
        else:
            x_range = np.arange(self.reader_info["xllcorner"], self.reader_info["xllcorner"] + self.reader_info["nrows"] * self.reader_info["dx"],
                                self.reader_info["dx"])
            y_range = np.arange(self.reader_info["yllcorner"], self.reader_info["yllcorner"] + self.reader_info["ncols"] * self.reader_info["dy"],
                                self.reader_info["dy"])
        self.x_mesh, self.y_mesh = np.meshgrid(x_range, y_range)
        self.y_mesh = np.flipud(self.y_mesh)  # To fit into the .asc format criteria

    def build_info(self):
        """
        Function that creates the internal data structure of the raster file
        """
        self.info["reader"]["filename"] = self.filename

    def add_z_info(self, z_coord):
        self.z_coord = z_coord

    def get_data(self) -> np.ndarray:
        if hasattr(self, "z_coord"):
            return self.get_xyz_data()
        else:
            return self.get_xy_data()

    def rebuild_x_y(self):
        x_range = np.arange(self.info["reader"]["xllcorner"], self.info["reader"]["nrows"] * self.info["reader"]["cellsize"],
                            self.info["reader"]["cellsize"])
        y_range = np.arange(self.info["reader"]["yllcorner"], self.info["reader"]["ncols"] * self.info["reader"]["cellsize"],
                            self.info["reader"]["cellsize"])
        self.x_mesh, self.y_mesh = np.meshgrid(x_range, y_range)

    def get_xy_data(self) -> np.ndarray:
        ndata = int(self.info["reader"]["nrows"] * self.info["reader"]["ncols"])
        self.xydata = np.zeros(shape=(ndata, 3))
        x_mesh_flatten = self.x_mesh.flatten()
        y_mesh_flatten = self.y_mesh.flatten()
        for id, data in enumerate(self.data.flatten()):
            self.xydata[id] = (x_mesh_flatten[id], y_mesh_flatten[id], data)
        self.xydata_computed = True
        return self.xydata

    def get_xyz_data(self) -> np.ndarray:
        assert hasattr(self, "z_coord"), "The z-coordinate of this raster file is not given"
        ndata = int(self.info["reader"]["nrows"] * self.info["reader"]["ncols"])
        self.flatten_data = np.zeros(shape=(ndata, 4))
        x_mesh_flatten = self.x_mesh.flatten()
        y_mesh_flatten = self.y_mesh.flatten()
        for id, data in enumerate(self.data.flatten()):
            self.flatten_data[id] = (x_mesh_flatten[id], y_mesh_flatten[id], self.z_coord, data)
        self.xydata_computed = True
        return self.flatten_data

    def to_csv(self, output_file, z_coord=None):
        """
        Function that writes the ratser data into a csv file
        :param output_file:
        :return:
        """
        print(f"Starting dump into {output_file}")
        if not self.xydata_computed:
            xydata = self.get_xy_data()
        else:
            xydata = self.xydata
        f = open(output_file, "w")
        if z_coord is not None:
            f.write(f"x,y,z,data\n")
        else:
            f.write(f"x,y,data\n")
        for data in xydata:
            if z_coord is not None:
                f.write(f"{data[0]},{data[1]},{z_coord},{data[2]}\n")
            else:
                f.write(f"{data[0]},{data[1]},{data[2]}\n")
        f.close()
        print(f"The data has been properly exported to the {output_file} file")

    def to_wsv(self, output_file):
        print(f"Starting dump into {output_file}")
        if not self.xydata_computed:
            xydata = self.dump_to_xydata()
        else:
            xydata = self.xydata
        f = open(output_file, "w")
        for data in xydata:
            f.write(f"{data[0]} {data[1]} {data[2]}\n")
        f.close()
        print(f"The data has been properly exported to the {output_file} file")

    def to_asc(self, output_file):
        print(f"Starting dump into {output_file}")
        file = open(output_file, "w")
        self.write_asc_header(file)
        self.write_asc_data(file)
        file.close()

    def write_asc_header(self, file):
        # assert isinstance(file, type(open)), "is not a correct file"
        for head in self.info:
            file.write(f"{head} {self.info[head]}\n")

    def write_asc_data(self, file):
        np.savetxt(file, self.data, fmt="%3.2f")

    def downsample_data(self, slice_factor=2):
        '''
        This module downsamples the data based on a constant stride in each direction
        :param slice_factor: factor to stride the matrix in each dimension

        :return: downsampled dataset
        '''
        self.data = self.data[0::slice_factor, 0::slice_factor]
        self.info["nrows"] = self.data.shape[0]
        self.info["ncols"] = self.data.shape[1]
        self.info["cellsize"] *= slice_factor
        RasterFileReader.rebuild_x_y(self)
        print("Data has been downsampled, the new settings are these:")
        print(self.info)

    def get_value_from_coord(self, x: float, y: float):
        """
        This method returns the raster value finding the nearest neighbour of a given x, y coordinate
        Args:
            x: Coordinate of the x-direction
            y: Coordinate of the y-direction

        Returns:
            Raster value at the given coordinate
        """
        # Define auxiliaty variables
        d_raster = self.reader_info["cellsize"]
        origin_x = self.reader_info["xllcorner"]
        origin_y = self.reader_info["yllcorner"]

        ix = int(np.floor((x - origin_x) / (1.001 * d_raster)))  # 1.001 value is used to avoid issues with
        # the floor function
        iy = int(np.floor((y - origin_y) / (1.001 * d_raster)))  # 1.001 value is used to avoid issues with
        # the floor function
        iy = int(self.reader_info["ncols"] - iy - 1)
        return self.data[iy, ix]



import unittest
import numpy.testing as nptest
import numpy as np
from pydelling.utils.geometry import Point


class PointCase(unittest.TestCase):
    def test_Point_init(self):
        p = Point([1,2,3])
        self.assertEqual(p.x, 1)
        self.assertEqual(p.y, 2)
        self.assertEqual(p.z, 3)

    def test_distance_point(self):
        p1: Point = Point([0, 0, 0])
        p2: Point = Point([1, 1, 0])
        self.assertEqual(p1.distance(p2), np.sqrt(2))


if __name__ == '__main__':
    unittest.main()



import functools

def set_run(func):
    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        value = func(self, *args, **kwargs)
        self.is_run = True
        return value
    return wrapper


import vtkmodules.vtkCommonDataModel

from pydelling.preprocessing import BasePreprocessing
from pydelling.config import config
from pydelling.utils.decorators import set_run
import open3d as o3d
import pandas as pd
import numpy as np
import logging
import vtk
from pathlib import Path

logger = logging.getLogger(__name__)


class STLFromPointCloud(BasePreprocessing):
    preprocessed_data: np.ndarray
    is_run: bool = False
    stl_mesh: o3d.geometry.TriangleMesh

    def __init__(self, data: pd.DataFrame = None, filename: str = None):
        super().__init__(data, filename)


    @set_run
    def run(self, method="ball_pivoting", *args, **kwargs):
        logger.info("Preprocessing cloud of points")
        assert self.is_data_ok(), "Data is not properly set-up"
        self.point_cloud = o3d.geometry.PointCloud()
        self.point_cloud.points = o3d.utility.Vector3dVector(self.data)
        self.point_cloud.estimate_normals(search_param=o3d.geometry.KDTreeSearchParamHybrid(radius=1.5,
                                                          max_nn=75))
        if config.stl_from_point_cloud.stl_strategy:
            method = config.stl_from_point_cloud.stl_strategy
        if method == 'ball_pivoting' or method == 'ball pivoting':
            if 'ball_radius' in kwargs:
                self.run_ball_pivoting(ball_radius=kwargs['ball_radius'], **kwargs)
            else:
                self.run_ball_pivoting()

        if method == 'poisson':
            self.run_poisson_reconstruction()

    def is_data_ok(self):
        if isinstance(self.data, pd.DataFrame):
            self.preprocessed_data = self.data.values
            return True
        elif isinstance(self.data, np.ndarray):
            self.preprocessed_data = self.data
            return True
        else:
            return False

    def visualize_point_cloud(self):
        o3d.visualization.draw_geometries([self.point_cloud])

    def visualize_stl_mesh(self, stl_mesh=None):
        if stl_mesh is not None:
            o3d.visualization.draw_geometries([stl_mesh])
        else:
            assert self.is_run, "Run the self.run() method before visualizing the results"
            o3d.visualization.draw_geometries([self.stl_mesh])

    @set_run
    def run_ball_pivoting(self, ball_radius=None, simplify_mesh=False, **kwargs):
        logger.info("Running the ball pivoting algorithm to recreate the stl surface")
        if not ball_radius and not config.stl_from_point_cloud.ball_radius:
            distances = self.point_cloud.compute_nearest_neighbor_distance()
            self.avg_distance = np.mean(distances)
            self.ball_radius = 3 * self.avg_distance * config.stl_from_point_cloud.buff if config.stl_from_point_cloud.buff else 1.0
            self.stl_mesh: o3d.geometry.TriangleMesh = o3d.geometry.TriangleMesh.create_from_point_cloud_ball_pivoting(
                pcd=self.point_cloud,
                radii=o3d.utility.DoubleVector([self.ball_radius, self.ball_radius * 2]),
                **config.stl_from_point_cloud.arguments,
            )
        else:
            self.ball_radius = ball_radius if ball_radius else config.stl_from_point_cloud.ball_radius
            self.stl_mesh: o3d.geometry.TriangleMesh = o3d.geometry.TriangleMesh.create_from_point_cloud_ball_pivoting(
                pcd=self.point_cloud,
                radii=o3d.utility.DoubleVector([self.ball_radius, self.ball_radius * 2]),
                **config.stl_from_point_cloud.arguments,
            )
        if simplify_mesh or config.stl_from_point_cloud.simplify_mesh:
            self.simplify_mesh()
        return self.stl_mesh


    def make_mesh_watertight(self,
                             hole_size=None,
                             filename=None,
                             write=True):
        """This method reads a ply mesh from an external file and converts it into a watertight mesh
        """
        self.to_ply(filename="temp.ply")
        logger.info(f"Making the mesh read from {filename} watertight")
        reader = vtk.vtkPLYReader()
        if filename is None:
            filename = Path.cwd() / "output/temp.ply"
        reader.SetFileName(str(filename))
        reader.Update()
        polydata = reader.GetOutput()
        fill = vtk.vtkFillHolesFilter()
        fill.SetInputData(polydata)
        hole_size = hole_size if hole_size else config.stl_from_point_cloud.watertight_mesh.hole_size if config.stl_from_point_cloud.watertight_mesh.hole_size else 1.0
        fill.SetHoleSize(hole_size)
        fill.Update()
        self.vtk_filled = fill.GetOutput()
        if write:
            writer = vtk.vtkPLYWriter()
            writer.SetInputData(self.vtk_filled)
            writer.SetFileName(f"temp-watertight.ply")
            writer.Write()
        self.stl_mesh = o3d.io.read_triangle_mesh("temp-watertight.ply")
        print(self.stl_mesh)
        # Path.unlink(Path("temp-watertight.ply"))
        # Path.unlink(filename)

        return fill.GetOutput()

    def reduce_mesh(self, target_value=100000):
        self.stl_mesh = self.stl_mesh.reduce_mesh(target_value)
        return self.stl_mesh


    def simplify_mesh(self, target_value=100000):
        self.stl_mesh.remove_degenerate_triangles()
        self.stl_mesh.remove_duplicated_triangles()
        self.stl_mesh.remove_duplicated_vertices()
        self.stl_mesh.remove_non_manifold_edges()
        return self.stl_mesh


    @set_run
    def run_poisson_reconstruction(self, depth=8, **kwargs):
        logger.info("Running the poisson reconstruction method to recreate the stl surface")
        self.point_cloud.orient_normals_towards_camera_location(self.point_cloud.get_center())
        self.point_cloud.normals = o3d.utility.Vector3dVector(- np.asarray(self.point_cloud.normals))
        self.stl_mesh = o3d.geometry.TriangleMesh.create_from_point_cloud_poisson(self.point_cloud,
                                                                                  depth=depth,
                                                                                  **kwargs if kwargs else {},
                                                                                  **config.stl_from_point_cloud.arguments,
                                                                                  )[0]
        return self.stl_mesh

    def to_ply(self, filename="reconstructed_mesh.ply"):
        """Writes the mesh in ply format
        """
        # assert self.is_run, "a mesh reconstruction technique needs to be run first"
        o3d.io.write_triangle_mesh(str(self.output_directory / filename), self.stl_mesh)
        logger.info(f"Writing to {str(self.output_directory / filename)}")

    def to_stl(self, filename="reconstructed_mesh.stl"):
        """Writes the mesh in stl format
        """
        # assert self.is_run, "a mesh reconstruction technique needs to be run first"
        self.stl_mesh = self.stl_mesh.compute_vertex_normals()
        logger.info(f"Writing to {str(self.output_directory / filename)}")
        o3d.io.write_triangle_mesh(str(self.output_directory / filename), self.stl_mesh)

    def read_stl(self, filename):
        """Reads a stl file and converts it into a o3d mesh"""
        self.stl_mesh = o3d.io.read_triangle_mesh(filename)

    def crop_bounding_box(self, bbox=None):
        """
        Uses original cloud of point bounding box to crop the STL mesh
        """
        if bbox is None:
            logger.info("Computing bounding box and cropping the mesh")
            bbox = self.point_cloud.get_axis_aligned_bounding_box()
        else:
            logger.info(f"Computing bounding box at {bbox} and cropping the mesh")
            bbox = o3d.geometry.AxisAlignedBoundingBox(min_bound=bbox[0], max_bound=bbox[1])

        self.stl_mesh = self.stl_mesh.crop(bbox)
        return self.stl_mesh

    @staticmethod
    def create_output_directory():
        """Creates an ./output folder on the working directory
        """
        Path(Path.cwd() / "output").mkdir(parents=True, exist_ok=True)

    @property
    def output_directory(self):
        self.create_output_directory()
        return Path.cwd() / "output"




import pandas as pd
from typing import List, Dict


class BaseReader:
    """Base class for the different reader subclasses"""
    filename: str
    data: Dict[str, pd.DataFrame]

    def read(self):
        """This function should read the data automatically from the config file"""
        pass



import numpy as np

from pydelling.config import config
from pydelling.readers.iGPReader.io import BaseWriter, iGPReader


class PflotranImplicitWriter(BaseWriter, iGPReader):
    """
    This class contains the function to write a mesh file in PFLOTRAN unstructured implicit format
    """

    def write_elements(self, export_file):
        export_file.write(f"ELEMENTS {self.mesh_info['n_elements']}\n")
        for element in self.elements:
            export_file.write(
                f"{config.globals.element_dict[len(element.coords)]} {' '.join(map(str, np.array(element.nodes) + 1))}\n")

    def write_nodes(self, export_file):
        export_file.write(f"VERTICES {self.mesh_info['n_nodes']}\n")
        for id, node in enumerate(self.nodes_output):
            export_file.write(f"{node[0]} {node[1]} {self.nodes[id][2]:5.5f}\n")

    def write_domain_hdf5(self, export_file):
        """
        Writes the domain info (i.e. grid info) into an hdf5 file
        :param export_file: HDF5 File object
        """
        if self.is_mesh_built:
            raise RuntimeError("Exporting built mesh into hdf5 still under development")
        else:
            #  Pre-process cell data
            cell_data = np.zeros(shape=(self.mesh_info["n_elements"], 9), dtype=np.int32)
            for id, cell in enumerate(self.elements):
                n_cell = len(cell)
                cell_data[id][0] = n_cell
                np.put(cell_data[id], range(1, len(cell) + 1), cell)
            # Pre-process node data
            node_data = np.array(self.nodes)
            domain_group = export_file.create_group("Domain")
            domain_group.create_dataset("Cells", data=cell_data)
            domain_group.create_dataset("Vertices", data=node_data)

    def write_regions_hdf5(self, export_file):
        def add_dataset(hdf5_file, group_name, dataset_type, data):
            temp_group = hdf5_file.create_group(group_name)
            temp_dataset = temp_group.create_dataset(dataset_type, data=data)

        def preprocess_region(dataset):
            """
            Pre-process a face dataset into PFLOTRAN hdf5 format
            :param dataset: input dataset
            :return: pre-processed face dataset
            """
            face_data = np.zeros(shape=(len(dataset), 4), dtype=np.int32)
            for id, face in enumerate(dataset):
                n_face = len(face)
                face_data[id][0] = n_face
                face = np.array(face)
                np.put(face_data[id], range(1, len(face)), face)
            return face_data

        if self.is_mesh_built:
            raise RuntimeError("Exporting built mesh into hdf5 still under development")
        else:
            # Pre-process datasets
            region_group = export_file.create_group("Regions")
            # Add materials
            #  Create 'all' material
            data_all = np.array(range(self.mesh_info["n_elements"])) + 1
            add_dataset(region_group, "all", "Cell Ids", data_all)
            for material_name in self.material_dict:
                add_dataset(region_group, material_name, "Cell Ids", self.material_dict[material_name])
            # Add regions
            for region_name in self.region_dict:
                preprocess_region_dataset = preprocess_region(self.region_dict[region_name]["elements"])
                add_dataset(region_group, region_name, "Vertex Ids", preprocess_region_dataset)



"""
Centroid file reader
"""
import numpy as np
from .BaseReader import BaseReader
import logging

logger = logging.getLogger(__name__)


class StructuredGridReader(BaseReader):
    def __init__(self, filename, var_pos=3, var_name="var", var_type=np.float32, centroid_pos=(0, 3), header=False):
        self.var_pos = None
        self.var = None
        self.var_name = None
        self.var_type = None
        self.centroid_pos = None
        self.header = None
        super().__init__(filename, var_pos=var_pos,
                         var_name=var_name,
                         var_type=var_type,
                         centroid_pos=centroid_pos,
                         header=header)

    def read_file(self, opened_file: BaseReader):
        temp_centroid = []
        temp_id = []
        for line in opened_file.readlines():
            data_row = line.split()
            temp_centroid.append(data_row[self.centroid_pos[0]:self.centroid_pos[1]])
            temp_id.append([data_row[self.var_pos]])
        self.data = np.array(temp_centroid, dtype=np.float32)
        self.var = np.array(temp_id, dtype=self.var_type)


    def read_header(self):
        """
        TODO: Add the header reader of the centroid file
        Reads the header of the file
        :return:
        """
        pass

    def get_data(self) -> np.ndarray:
        """
        Outputs the data
        :return: np.ndarray object containing centroid information and variable output
        """
        return np.hstack((self.data, self.var))

    def build_info(self):
        """
        Generates a dictionary containing the basic info of the read data
        :return:
        """
        self.info["reader"] = {"n_cells": self.data.shape[0],
                     "filename": self.filename,
                     "var_name": self.var_name,
                     "var_position": self.var_pos}

    def dump_to_csv(self, output_file, delimiter=","):
        """
        Writes the data into a csv file
        :param output_file:
        :return:
        """
        print(f"Starting dump into {output_file}")
        np.savetxt(output_file, self.get_data(), delimiter=delimiter)
        print(f"The data has been properly exported to the {output_file} file")




"""
This class implements a basic interface for Interpolation classes
"""
import numpy as np
import h5py
import os
from ..writers.BaseWriter import BaseWriter
import seaborn as sns
import pandas as pd
import logging
from pydelling.utils.decorators import set_run
from pydelling.config import config

logger = logging.getLogger(__name__)


class BaseInterpolator:
    is_run: bool = False
    has_regular_mesh: bool = False

    def __init__(self,
                 interpolation_data=None,
                 mesh_data=None):
        self.data: np.ndarray = []
        self.mesh: np.ndarray = []
        self.info = {"interpolation": {}}
        self.interpolated_data: np.ndarray = []
        self.filename = None
        if interpolation_data is not None:
            self.add_data(data=interpolation_data)
        if mesh_data is not None:
            self.add_mesh(data=mesh_data)

    def add_data(self, data):
        """
        Add a dataset that will be used to interpolate
        :return:
        """
        if len(self.data) == 0:
            self.data = np.array(data)
        else:
            self.data = np.vstack((self.data, data))

    def add_mesh(self, data, id_index=3):
        """
        Add the set of points on which interpolation will be performed
        :return:
        """
        if data.shape[1] > 3:
            temp_id_data = data[:, id_index]
            temp_data = data[:, 0:3]
        else:
            temp_data = data

        if not self.mesh:
            self.mesh = np.array(temp_data)
            if data.shape[1] > 3:
                self.id_data = np.array(temp_id_data)
        else:
            self.mesh = np.vstack((self.mesh, temp_data))
            if data.shape[1] > 3:
                self.id_data = np.vstack((self.id_data, temp_id_data))

    @set_run
    def run(self):
        """
        Runs the interpolation algorithm
        Returns:
            A numpy array containing the values of the interpolated variable on the mesh centroids
        """
        self.interpolated_data = self.mesh[:, 0]
        return self.interpolated_data

    def get_data(self):
        """
        Returns interpolated data
        :return:
        """
        return self.interpolated_data

    def dump_to_hdf5(self, filename=None, var_name=None, data=None):
        """
        Dumps the data into HDF5 format
        :return:
        """
        if data is None:
            data = self.interpolated_data
        if filename is not None:
            self.filename = filename
        if not os.path.exists(filename):
            tempfile = h5py.File(filename, "w")
            tempfile.close()
        with h5py.File(filename, "r+") as tempfile:
            tempfile.create_dataset(var_name, data=data)

    def dump_to_csv(self, filename=None, **kwargs):
        temp_array = np.reshape(self.interpolated_data, (self.interpolated_data.shape[0], 1))
        temp_array = np.concatenate((self.mesh, temp_array), axis=1)
        np.savetxt(filename, temp_array, **kwargs)
        if config.general.verbose:
            print(f"Data has been dumped into {filename}")

        # return np.concatenate((self.mesh, temp_array), axis=1)

    def wipe_data(self):
        """Wipes data structure
        """
        self.data = []
        self.mesh = []
        self.interpolated_data = []

    def write_data(self, writer_class=BaseWriter, filename=None, **kwargs):
        base_writer = writer_class(filename=filename, data=self.get_data(), info=self.info, **kwargs)
        base_writer.run(filename=filename)

    def remove_output_file(self, writer_class=BaseWriter, filename=None, **kwargs):
        base_writer = writer_class(filename=filename, **kwargs)
        base_writer.remove_output_file(filename)

    def get_minmax_coords(self):
        self.data_xmin = np.min(self.data[:, 0])
        self.data_xmax = np.max(self.data[:, 0])
        self.data_ymin = np.min(self.data[:, 1])
        self.data_ymax = np.max(self.data[:, 1])
        self.info["interpolation"]["x_min"] = self.data_xmin
        self.info["interpolation"]["x_max"] = self.data_xmax
        self.info["interpolation"]["y_min"] = self.data_ymin
        self.info["interpolation"]["y_max"] = self.data_ymax

    def create_regular_mesh(self, n_x, n_y, dilatation_factor=1.0):
        """Create an inner regular mesh"""
        self.has_regular_mesh = True
        self.mesh = []
        self.get_minmax_coords()
        dx = abs(self.data_xmax - self.data_xmin) / n_x
        dx_dil = dx * dilatation_factor
        dx_correction = (dx_dil - dx) / 2.0

        dy = abs(self.data_ymax - self.data_ymin) / n_y
        dy_dil = dy * dilatation_factor
        dy_correction = (dy_dil - dy) / 2.0

        self.info["interpolation"].update({"n_x": n_x,
                                           "n_y": n_y,
                                           "dilatation_factor": dilatation_factor,
                                           "type": "regular_mesh",
                                           "d_x": dx_dil,
                                           "d_y": dy_dil,
                                           })

        linspace_x = np.linspace(self.data_xmin - dx_correction, self.data_xmax + dx_correction, n_x)
        linspace_y = np.linspace(self.data_ymin - dy_correction, self.data_ymax + dy_correction, n_y)
        grid_x, grid_y = np.meshgrid(linspace_x, linspace_y)
        self.mesh = np.hstack((grid_x.reshape((grid_x.size, 1)), grid_y.reshape((grid_y.size, 1))))

    def describe(self, write_to_file=None, plots=True):
        assert self.is_run, "The interpolator has not been run"
        temp_df = pd.DataFrame(self.interpolated_data)
        logger.info("Describing the interpolated data")
        print(temp_df.describe())
        if write_to_file:
            if type(write_to_file) is str:
                temp_df.describe().to_csv(write_to_file)
            else:
                temp_df.describe().to_csv("interpolated_data-description.csv")
            logger.info("Writing the description to file")
        if plots:
            logger.info("Plotting data")
            import matplotlib.pyplot as plt
            histogram_plot = sns.kdeplot(x=self.interpolated_data)
            plt.show()







from pydelling.paraview_processor.filters import BaseFilter
try:
    from paraview.simple import *
except:
    pass




class IntegrateVariablesFilter(BaseFilter):
    """
    This class implements the Paraview Integrate Variables filter
    """
    filter_type: str = "Integrate_variables"
    counter: int = 0

    def __init__(self, input_filter, name, divide_cell_data_by_volume=False):
        super().__init__(name=name)
        IntegrateVariablesFilter.counter += 1
        self.filter = IntegrateVariables(Input=input_filter)
        self.set_divide_cell_data_by_volume(divide_cell_data_by_volume)

    def set_divide_cell_data_by_volume(self, value):
        self.filter.DivideCellDataByVolume = value




import numpy as np
from scipy.spatial.qhull import ConvexHull
from shapely.geometry import Polygon

# Useful geometrical functions
def normal_vector(points):
    """Computes normal vector of a co-planar set of points (vertices of a polygon)"""
    assert len(points) >= 3, "Incorrect number of points, more are needed to form a plane"
    v1 = points[1] - points[0]
    v2 = points[2] - points[0]
    vn = np.cross(v1, v2)
    vn_norm = np.linalg.norm(vn)
    return vn / vn_norm


def line_plane_intersection(line_points=[], plane_points=[]):
    # Compute plane normal
    # line points
    p0 = line_points[0]
    p1 = line_points[1]
    u = p1 - p0
    # plane points
    v0 = plane_points[0]
    vn = normal_vector(plane_points)
    s = (np.dot(vn, (v0 - p0)) / np.dot(vn, u))
    intersection_point = p0 + s * u
    return intersection_point


def line_plane_perpendicular_intersection(line_point=None, plane_points=[]):
    # Compute plane normal
    # line points
    p0 = line_point
    # plane points
    v0 = plane_points[0]
    vn = normal_vector(plane_points)
    s = (np.dot(vn, (v0 - p0)) / np.dot(vn, vn))
    intersection_point = p0 + s * vn
    return intersection_point


def compute_polygon_area(points) -> float:
    """
    This function computes the area of a 3D planar polygon
    :return: Area of a 3D planar polygon
    """

    if len(points) < 3:
        return 0.0

    # convex_hull = Polygon(points)
    # return convex_hull.convex_hull.area

    # Compute normal
    points = order_points_clockwise(points)
    vn = normal_vector(points)
    temp_area_v = np.zeros(shape=3)
    projected_area = 0.0
    n_coords = len(points)
    for id, point in enumerate(points):
        # Set-up variables
        v1 = points[id % n_coords]  # Pv1, assuming P=(0,0,0)
        v2 = points[(id + 1) % n_coords]  # Pv2, assuming P=(0,0,0)
        # Compute area
        id_area_v = np.cross(v1, v2)
        projected_area_id = np.dot(vn, id_area_v) / 2.0  # area of small triangle of the face
        projected_area += projected_area_id
    return projected_area


def order_points_clockwise(points: np.ndarray) -> list:
    """Orders set of 3D coplanar points clockwise"""
    # Compute normal vector
    vn = normal_vector(points)
    # Compute centroid
    centroid = np.mean(points, axis=0)
    # Project points on normal plane
    v1 = points[0] - centroid
    v1 = v1 / np.linalg.norm(v1)
    v2 = np.cross(vn, v1)
    v2 = v2 / np.linalg.norm(v2)
    projected_points = np.zeros(shape=(len(points), 3))
    for id, point in enumerate(points):
        projected_points[id] = [np.dot(v1, point), np.dot(v2, point), 0.0]
    # Order points clockwise, compute angles
    center = np.mean(projected_points, axis=0)
    angles = np.zeros(shape=len(projected_points))
    for id, point in enumerate(projected_points):
        angles[id] = np.arctan2(point[1] - center[1], point[0] - center[0])
    # Sort points
    sorted_points = [points[id] for id in np.argsort(angles)]
    return sorted_points


def filter_unique_points(points: list or np.ndarray, tolerance: float = 1e-4) -> list:
    """Filters out duplicate points"""
    # Convert to numpy array
    if isinstance(points, list):
        points = np.array(points)
    # Filter unique points
    unique_points = []
    for point in points:
        if len(unique_points) == 0:
            unique_points.append(point)
        uniqueness = True
        for unique_point in unique_points:
            if np.linalg.norm(point - unique_point) < tolerance:
                uniqueness = False
        if uniqueness:
            unique_points.append(point)
    return unique_points




import numpy as np


# Useful geometrical functions
def normal_vector(points):
    assert len(points) >= 3, "Incorrect number of points, more are needed to form a plane"
    v1 = points[1] - points[0]
    v2 = points[2] - points[0]
    vn = np.cross(v1, v2)
    vn_norm = np.linalg.norm(vn)
    return vn / vn_norm


def line_plane_intersection(line_points=[], plane_points=[]):
    # Compute plane normal
    # line points
    p0 = line_points[0]
    p1 = line_points[1]
    u = p1 - p0
    # plane points
    v0 = plane_points[0]
    vn = normal_vector(plane_points)
    s = (np.dot(vn, (v0 - p0)) / np.dot(vn, u))
    intersection_point = p0 + s * u
    return intersection_point


def line_plane_perpendicular_intersection(line_point=None, plane_points=[]):
    # Compute plane normal
    # line points
    p0 = line_point
    # plane points
    v0 = plane_points[0]
    vn = normal_vector(plane_points)
    s = (np.dot(vn, (v0 - p0)) / np.dot(vn, vn))
    intersection_point = p0 + s * vn
    return intersection_point



from pydelling.paraview_processor.filters import BaseFilter
try:
    from paraview.simple import *
except:
    pass


class AppendArcLengthFilter(BaseFilter):
    """
    This class implements the Append Arc-Length filter
    """
    filter_type: str = "Append_arc_length"
    counter: int = 0

    def __init__(self, input_filter, name):
        super().__init__(name=name)
        AppendArcLengthFilter.counter += 1
        self.filter = AppendArcLength(Input=input_filter)


from __future__ import annotations
from .BasePrimitive import BasePrimitive
from .Point import Point
from .Line import Line
from .Vector import Vector
import numpy as np
from typing import List


class Plane(BasePrimitive):
    def __init__(self, point: Point or List or np.ndarray, normal: List or Vector or np.ndarray):
        self.p = Point(point)
        self.n = Vector(normal)

    def __repr__(self):
        return f"Plane(point:{self.p}, normal:{self.n})"

    def __str__(self):
        return f"Plane(point:{self.p}, normal:{self.n})"

    def intersect(self, primitive: BasePrimitive):
        """Returns the intersection of this plane with the given primitive"""
        from .intersections import intersect_plane_plane, intersect_plane_line, intersect_plane_segment
        if primitive.__class__.__name__ == "Plane":
            return intersect_plane_plane(plane_1=self, plane_2=primitive)
        elif primitive.__class__.__name__ == "Line":
            return intersect_plane_line(plane=self, line=primitive)
        elif primitive.__class__.__name__ == 'Segment':
            return intersect_plane_segment(plane=self, segment=primitive)

        else:
            raise NotImplementedError(f"Intersection with {type(primitive)} is not implemented")

    def is_parallel(self, plane: Plane):
        """Returns True if this plane is parallel to the given plane"""
        if np.isclose(np.dot(self.n, plane.n), 1):
            return True
        if np.isclose(np.dot(self.n, plane.n), -1):
            return True




import os

from pydelling.readers.iGPReader.io import iGPReader
from pydelling.readers.iGPReader.utils.geometry_utils import *
from pathlib import Path


class CsvWriter(iGPReader):
    def write_csv_cells(self):
        if not self.output_folder:
            file_csv = open(f"{self.project_name}_cell.csv", "w")
        else:
            file_csv = open(Path.cwd() / f"output/{self.project_name}_cell.csv", "w")
        file_csv.write("X,Y,Z,V\n")
        for element in self.elements:
            file_csv.write(f"{element.centroid[0]},{element.centroid[1]},{element.centroid[2]},{element.volume}\n")
        file_csv.close()

    def write_csv_connection(self):
        if self.output_folder is None:
            file_csv = open(f"{self.project_name}_connections.csv", "w")
        else:
            file_csv = open(os.path.join(self.output_folder, f"{self.project_name}_connections.csv"), "w")
        file_csv.write("X,Y,Z,A\n")
        for element in self.connections:
            prime_element = element[0]
            for connected_element in element[1]:
                face_id = element[1][connected_element]
                face_obj = self.elements[prime_element].faces[face_id]
                face_nodes = face_obj.coords
                # Compute line-face intersection
                line_points = np.array([self.elements[prime_element].centroid_coords,
                                        self.elements[connected_element].centroid_coords])
                intersection_point = line_plane_intersection(line_points=line_points,
                                                                      plane_points=face_nodes)
                # compute connection centroid and area
                conn_area = face_obj.area
                conn_centroid = face_obj.centroid
                # export_file.write(f"{prime_element + 1} {connected_element + 1} {conn_centroid[0]:1.4e} {conn_centroid[1]:1.4e} {conn_centroid[2]:1.4e} {conn_area:1.4e}\n")
                file_csv.write(f"{intersection_point[0]},{intersection_point[1]},{intersection_point[2]},{conn_area}\n")
        file_csv.close()



import unittest
import numpy.testing as nptest
import numpy as np
from pydelling.utils.geometry import Line
from pydelling.preprocessing import DfnPreprocessor, MeshPreprocessor
from pydelling.preprocessing.dfn_preprocessor import DfnUpscaler

class IntersectionCase(unittest.TestCase):
    def test_full_intersection(self):
        dfn_preprocessor = DfnPreprocessor()
        dfn_preprocessor.add_fracture(
            x=0.1, y=0.1, z=0.0, dip=0, dip_dir=0, size=1.2
        )
        mesh_preprocessor = MeshPreprocessor()
        mesh_preprocessor.add_tetrahedra(
            node_ids=np.array([0, 1, 2, 3]),
            node_coords=[
                np.array([0.0, 0.0, -0.5]),
                np.array([1.0, 0.0, -0.5]),
                np.array([0.0, 1.0, -0.5]),
                np.array([0.0, 0.0, 0.5]),
            ],
        )
        # Intersect fracture with mesh
        intersected_points = mesh_preprocessor.elements[0].intersect_faces_with_plane(
            dfn_preprocessor[0].plane)
        solution = [
            np.array([0.5, 0.0, 0.0]),
            np.array([0.0, 0.0, 0.0]),
            np.array([0.0, 0.5, 0.0]),
        ]
        nptest.assert_array_almost_equal(intersected_points, solution)

    def test_full_intersection_rotated(self):
        dfn_preprocessor = DfnPreprocessor()
        dfn_preprocessor.add_fracture(
            x=0.1, y=0.1, z=0.0, dip=20, dip_dir=50, size=1.2
        )
        mesh_preprocessor = MeshPreprocessor()
        mesh_preprocessor.add_tetrahedra(
            node_ids=np.array([0, 1, 2, 3]),
            node_coords=[
                np.array([0.0, 0.0, -0.5]),
                np.array([1.0, 0.0, -0.5]),
                np.array([0.0, 1.0, -0.5]),
                np.array([0.0, 0.0, 0.5]),
            ],
        )
        # Intersect fracture with mesh
        intersected_points = mesh_preprocessor.elements[0].intersect_faces_with_plane(
            dfn_preprocessor[0].plane)
        solution = [
            np.array([0.62220399, 0.0, -0.12220399]),
            np.array([0.0, 0.0, 5.12772932e-02]),
            np.array([0.0, 5.85765892e-01, -8.57658923e-02]),
        ]
        nptest.assert_array_almost_equal(intersected_points, solution)

    def test_partial_intersection_1(self):

        dfn_preprocessor = DfnPreprocessor()
        dfn_preprocessor.add_fracture(
            x=0.1, y=0.1, z=0.5, dip=90, dip_dir=90, size=1.5
        )
        mesh_preprocessor = MeshPreprocessor()
        mesh_preprocessor.add_tetrahedra(
            node_ids=np.array([0, 1, 2, 3]),
            node_coords=[
                np.array([0.0, 0.0, -0.5]),
                np.array([1.0, 0.0, -0.5]),
                np.array([0.0, 1.0, -0.5]),
                np.array([0.0, 0.0, 0.5]),
            ],
        )
        # Intersect fracture with mesh
        intersected_points = mesh_preprocessor.elements[0].intersect_faces_with_plane(
            dfn_preprocessor[0].plane)

        solution = [
            np.array([0.1, 0.0, 0.4]),
            np.array([0.1, 0.0, -0.5]),
            np.array([0.1, 0.9, -0.5]),
        ]
        nptest.assert_array_almost_equal(intersected_points, solution)


    def test_partial_intersection_2(self):

        dfn_preprocessor = DfnPreprocessor()
        dfn_preprocessor.add_fracture(
            x=0.1, y=0.1, z=0.5, dip=75, dip_dir=85, size=1.5
        )
        mesh_preprocessor = MeshPreprocessor()
        mesh_preprocessor.add_tetrahedra(
            node_ids=np.array([0, 1, 2, 3]),
            node_coords=[
                np.array([0.0, 0.0, -0.5]),
                np.array([1.0, 0.0, -0.5]),
                np.array([0.0, 1.0, -0.5]),
                np.array([0.0, 0.0, 0.5]),
            ],
        )
        # Intersect fracture with mesh
        intersected_points = mesh_preprocessor.elements[0].intersect_faces_with_plane(
            dfn_preprocessor[0].plane)

        solution = [
            np.array([0.14876171, 0.0, 0.35123829]),
            np.array([0.37772158, 0.0, -0.5]),
            np.array([0.31805952, 0.68194048, -0.5]),
        ]
        nptest.assert_array_almost_equal(intersected_points, solution)


    def test_partial_intersection_3(self):
        dfn_preprocessor = DfnPreprocessor()
        dfn_preprocessor.add_fracture(
            x=0.1, y=0.1, z=0.3, dip=75, dip_dir=65, size=1.3
        )
        mesh_preprocessor = MeshPreprocessor()
        mesh_preprocessor.add_tetrahedra(
            node_ids=np.array([0, 1, 2, 3]),
            node_coords=[
                np.array([0.0, 0.0, -0.5]),
                np.array([1.0, 0.0, -0.5]),
                np.array([0.0, 1.0, -0.5]),
                np.array([0.0, 0.0, 0.5]),
            ],
        )
        # Intersect fracture with mesh
        intersected_points = mesh_preprocessor.elements[0].intersect_faces_with_plane(
            dfn_preprocessor[0].plane)

        solution = [
            np.array([0.12422918, 0.0, 0.37577082]),
            np.array([0.38315014, 0.0, -0.5]),
            np.array([0.0, 5.12725439e-01, -1.27254385e-02]),
            np.array([0.0, 0.82166813, -0.5]),
        ]
        nptest.assert_array_almost_equal(intersected_points, solution)


    def test_partial_intersection_one_corner_inside(self):
        dfn_preprocessor = DfnPreprocessor()
        dfn_preprocessor.add_fracture(
            x=0.1, y=0.4, z=0.2, dip=75, dip_dir=65, size=0.7
        )
        mesh_preprocessor = MeshPreprocessor()
        mesh_preprocessor.add_tetrahedra(
            node_ids=np.array([0, 1, 2, 3]),
            node_coords=[
                np.array([0.0, 0.0, -0.5]),
                np.array([1.0, 0.0, -0.5]),
                np.array([0.0, 1.0, -0.5]),
                np.array([0.0, 0.0, 0.5]),
            ],
        )
        # Intersect fracture with mesh
        intersected_points = mesh_preprocessor.elements[0].intersect_with_fracture(
            dfn_preprocessor[0])

        solution = [
            np.array([0.16664204, 0.471432, -0.13807404]),
            np.array([0.25949777, 0.08819276, 0.15230947]),
            np.array([0.33001579, 0.12107585, -0.13807404]),
        ]
        nptest.assert_array_almost_equal(intersected_points, solution)


    def test_no_intersection_fracture_inside(self):
        dfn_preprocessor = DfnPreprocessor()
        dfn_preprocessor.add_fracture(
            x=0.2, y=0.2, z=0.0, dip=60, dip_dir=45, size=0.2
        )
        mesh_preprocessor = MeshPreprocessor()
        mesh_preprocessor.add_tetrahedra(
            node_ids=np.array([0, 1, 2, 3]),
            node_coords=[
                np.array([0.0, 0.0, -0.5]),
                np.array([1.0, 0.0, -0.5]),
                np.array([0.0, 1.0, -0.5]),
                np.array([0.0, 0.0, 0.5]),
            ],
        )
        # Intersect fracture with mesh
        intersected_points = mesh_preprocessor.elements[0].intersect_with_fracture(
            dfn_preprocessor[0])

        solution = [
            np.array([0.30606602, 0.16464466, -0.08660254]),
            np.array([0.16464466, 0.30606602, -0.08660254]),
            np.array([0.09393398, 0.23535534, 0.08660254]),
            np.array([0.23535534, 0.09393398, 0.08660254]),
        ]
        nptest.assert_array_almost_equal(intersected_points, solution)

    def test_intersect_hexahedra_full(self):
        dfn_preprocessor = DfnPreprocessor()
        dfn_preprocessor.add_fracture(
            x=0.0, y=0.0, z=0.0, dip=1, dip_dir=0, size=2.0
        )
        mesh_preprocessor = MeshPreprocessor()
        mesh_preprocessor.add_hexahedra(
            node_ids=np.array([0, 1, 2, 3, 4, 5, 6, 7]),
            node_coords=[
                np.array([-0.5, -0.5, -0.5]),
                np.array([0.5, -0.5, -0.5]),
                np.array([0.5, 0.5, -0.5]),
                np.array([-0.5, 0.5, -0.5]),
                np.array([-0.5, -0.5, 0.5]),
                np.array([0.5, -0.5, 0.5]),
                np.array([0.5, 0.5, 0.5]),
                np.array([-0.5, 0.5, 0.5])
            ],
        )

        intersections = mesh_preprocessor.elements[0].intersect_with_fracture(
            dfn_preprocessor[0])
        solution = [
            np.array([-0.5, -0.5, 0.00872753]),
            np.array([-0.5, 0.5, -0.00872753]),
            np.array([0.5, -0.5, 0.00872753]),
            np.array([0.5, 0.5, -0.00872753]),
        ]
        np.testing.assert_array_almost_equal(intersections, solution)





if __name__ == "__main__":
    unittest.main()



import unittest

import numpy as np

from pydelling.preprocessing.mesh_preprocessor import MeshPreprocessor
from pydelling.readers.FemReader import FemReader
from pydelling.utils import test_data_path
from pydelling.preprocessing import DfnPreprocessor


class TestMeshPreprocessor(unittest.TestCase):
    def test_read_data(self):
        pass

    def test_intersection(self):
        fem_reader = FemReader(test_data_path() / "fem_reader_data.fem")
        dfn_test = DfnPreprocessor()
        # dfn_test.add_fracture(dip=45, dip_dir=0, x=5.0, y=5.0, z=-0.5, size=1)
        dfn_test.add_fracture(dip=45, dip_dir=0, x=5.0, y=5.0, z=-0.5, size=1.5)
        dfn_test.to_obj(filename='./test.obj')
        fem_reader.to_vtk(filename='./test.vtk')
        # fem_reader.find_the_intersection_between_fracture_and_mesh(dfn_test[0])
        fem_reader.find_intersection_points_between_fracture_and_mesh(dfn_test[0])

    def test_read_pyramid(self):
        mesh_preprocessor = MeshPreprocessor()
        mesh_preprocessor.add_pyramid(node_ids=np.array([0, 1, 2, 3, 4]),
                                      node_coords=[np.array([0.0, 0.0, 0.0]), np.array([1.0, 0.0, 0.0]),
                                                   np.array([1.0, 1.0, 0.0]), np.array([0.0, 1.0, 0.0]),
                                                   np.array([0.0, 0.0, 1.0])])
        self.assertEqual(mesh_preprocessor.n_nodes, 5)


    def test_normal_vector(self):
        pass



if __name__ == '__main__':
    unittest.main()


"""
Base interface for a reader class
"""
import numpy as np
import logging
from pydelling.readers import BaseReader
logger = logging.getLogger(__name__)
from pydelling.config import config
import logging
from pathlib import Path
import pandas as pd
logger = logging.getLogger(__name__)
from linecache import getline, getlines


class ConnectFlowReader(BaseReader):
    def __init__(self, filename=None):
        self.filename = Path(filename) if filename else Path(config.open_foam_reader.filename)
        logger.info(f"Reading ConnectFlow mesh file from {self.filename}")
        super().__init__(filename=self.filename)

    def open_file(self, filename):
        filename_string = str(filename)
        header = getline(filename_string, 1).split()
        self.n_nodes = int(header[1])
        self.n_elements = int(header[3])
        self.n_mat = int(header[5])
        with open(filename, "r") as opened_file:
            _nodes = [line.rstrip().split() for line in opened_file.readlines()[2: 2 + self.n_nodes - 1]]
            self._mesh_nodes = pd.DataFrame(np.array(_nodes).astype(np.float), columns=["index", "x", "y", "z"])
            self._mesh_nodes = self.mesh_nodes.set_index("index")
        self.build_info()

    @property
    def mesh_nodes(self) -> pd.DataFrame:
        assert hasattr(self, "_mesh_nodes"), "Mesh has not been properly assigned"
        return self._mesh_nodes


    def get_data(self) -> np.ndarray:
        """
        Outputs the read data
        :return:
        """
        return np.array(0)


    def build_info(self):
        """
        Generates a dictionary containing the basic info of the read data
        :return:
        """
        min_x = self._mesh_nodes["x"].min()
        max_x = self._mesh_nodes["x"].max()
        min_y = self._mesh_nodes["y"].min()
        max_y = self._mesh_nodes["y"].max()
        min_z = self._mesh_nodes["z"].min()
        max_z = self._mesh_nodes["z"].max()
        self.info = {
            "bounds": {
                "x": [min_x, max_x],
                "y": [min_y, max_y],
                "z": [min_z, max_z],
            }
        }
        self.info.update({
            "span": {
                "x": self.info["bounds"]["x"][1] - self.info["bounds"]["x"][0],
                "y": self.info["bounds"]["y"][1] - self.info["bounds"]["y"][0],
                "z": self.info["bounds"]["z"][1] - self.info["bounds"]["z"][0],
            }
        })

    def get_bounds(self):
        """
        Returns the x, y and z axis bounds
        Returns:
            A list containing the bounds
        """
        return self.info["bounds"]


    def get_span(self):
        """
        Returns the x, y and z axis span
        Returns:
            A list containing the span
        """
        return self.info["span"]


import numpy as np
from .BaseReader import BaseReader
try:
    from paraview.simple import *
    from paraview.vtk.numpy_interface import dataset_adapter as dsa
    from paraview import servermanager as sm
except:
    pass


class VtkReader(BaseReader):
    """
    This class provides the framework to read data from a VTK file and do different postprocessing steps
    """
    current_array: None
    calculator: None

    def read_file(self, opened_file):
        self.vtk_file = LegacyVTKReader(FileNames=self.filename)
        self.current_array = self.vtk_file

    @property
    def values(self):
        _vtk_object = sm.Fetch(self.current_array)
        _vtk_object = dsa.WrapDataObject(_vtk_object)
        return _vtk_object.PointData

    @property
    def keys(self):
        return self.current_array

    @property
    def data_keys(self):
        _vtk_object = sm.Fetch(self.vtk_file)
        return self.vtk_file.PointData.keys()

    @property
    def data_values(self):
        _vtk_object = sm.Fetch(self.vtk_file)
        _vtk_object = dsa.WrapDataObject(_vtk_object)
        return _vtk_object.PointData

    def add_calculator(self, input=None, function=''):
        """
        Adds a calculator filter to a dataset
        Returns:
            The Calculator object
        """
        input = input if input else self.current_array
        self.calculator = Calculator(Input=input)
        self.calculator.Function = function

        self.current_array = self.calculator
        return self.current_array


import pandas as pd

from pydelling.webapps.components import BaseComponent
from pydelling.webapps import WebAppRunner
from pydelling.config import config
import streamlit as st
from pydelling.readers import VTKMeshReader, FemReader, SmeshReader
from pathlib import Path
from translate import Translator


class InputComponent(BaseComponent):
    """Creates a streamlit block in which the user can drag and drop a file and the resulting object is returned"""
    def __init__(self, input_type: str, key: str, webapp: WebAppRunner=None, lang=None):
        self.key = key
        if not f'{self.key}_done' in st.session_state:
            st.session_state[f'{self.key}_done'] = False
        self.translate = False

        self.input_type = input_type
        if lang is not None:
            self.translator = Translator(to_lang=lang)
            self.translate = True

        super().__init__(webapp)


    def run(self):
        """Runs the component"""
        if self.input_type == 'csv':
            self.run_csv_input()
        elif self.input_type == 'vtk':
            self.run_vtk_input()
        elif self.input_type == 'fem':
            self.run_fem_input()
        elif self.input_type == 'smesh':
            self.run_smesh_input()

    def run_csv_input(self):
        with st.form(f"{self.key}_form"):
            st.write("Drop your input data (the data that contains the information to interpolate) in the box below")
            data = st.file_uploader("Upload your data", type=config.globals.webapp.supported_input_data)
            st.markdown('**Advanced upload options:**')
            cols = st.columns([1, 6, 6])
            with cols[0]:
                sep = st.text_input('Separator', ',')
            whitespace_delimiter = st.checkbox('Whitespace delimiter', value=False)
            # Every form must have a submit button.
            submitted = st.form_submit_button("Upload")
            if submitted:
                with(st.spinner(f'Loading data')):
                    if whitespace_delimiter:
                        read_data = pd.read_csv(data, delim_whitespace=whitespace_delimiter)
                    else:
                        read_data = pd.read_csv(data, sep=sep)
                    st.dataframe(read_data.head())
                    self._value = read_data
                    st.session_state[f'{self.key}_done'] = True
                    if not self.key in st.session_state:
                        st.session_state[f'{self.key}'] = read_data


    def run_vtk_input(self):
        desc = "Drop your mesh/centroid data (the mesh/location you want to interpolate into) in the box below"
        format_desc = 'This data must be in VTK format'
        upload_desc = 'Upload your data'
        upload_button = 'Upload'
        if self.translate:
            desc = self.translator.translate(desc)
            format_desc = self.translator.translate(format_desc)
            upload_desc = self.translator.translate(upload_desc)
            upload_button = self.translator.translate(upload_button)

        with st.form(f'{self.key}_form'):
            st.write(desc)
            st.markdown(format_desc)
            data = st.file_uploader(upload_desc, type='vtk')
            new_submit = st.form_submit_button(upload_button)
            if new_submit:
                # Write file to temp directory
                with open('temp.vtk', 'wb') as f:
                    f.write(data.getbuffer())
                vtk_mesh = VTKMeshReader('temp.vtk', st_file=True)
                Path('temp.vtk').unlink()
                st.session_state[f'{self.key}'] = vtk_mesh
                st.session_state[f'{self.key}_done'] = True


    def run_fem_input(self):
        desc = "Drop your mesh/centroid data (the mesh/location you want to interpolate into) in the box below"
        format_desc = 'This data must be in FEM ASCII format'
        upload_desc = 'Upload your data'
        upload_button = 'Upload'
        if self.translate:
            desc = self.translator.translate(desc)
            format_desc = self.translator.translate(format_desc)
            upload_desc = self.translator.translate(upload_desc)
            upload_button = self.translator.translate(upload_button)

        with st.form(f'{self.key}_form'):
            st.write(desc)
            st.markdown(format_desc)
            data = st.file_uploader(upload_desc, type='fem')
            new_submit = st.form_submit_button(upload_button)
            if new_submit:
                # Write file to temp directory
                with open('temp.fem', 'wb') as f:
                    f.write(data.getbuffer())
                vtk_mesh = FemReader('temp.fem', st_file=True)
                Path('temp.fem').unlink()
                st.session_state[f'{self.key}'] = vtk_mesh
                st.session_state[f'{self.key}_done'] = True


    def run_smesh_input(self):
        desc = "Drop your mesh/centroid data (the mesh/location you want to interpolate into) in the box below"
        format_desc = 'This data must be in SMesh ASCII format'
        upload_desc = 'Upload your data'
        upload_button = 'Upload'
        if self.translate:
            desc = self.translator.translate(desc)
            format_desc = self.translator.translate(format_desc)
            upload_desc = self.translator.translate(upload_desc)
            upload_button = self.translator.translate(upload_button)
        with st.form(f'{self.key}_form'):
            st.write(desc)
            st.markdown(format_desc)
            data = st.file_uploader(upload_desc, type='smesh')
            new_submit = st.form_submit_button(upload_button)
            if new_submit:
                # Write file to temp directory
                with open('temp.smesh', 'wb') as f:
                    f.write(data.getbuffer())
                vtk_mesh = SmeshReader('temp.smesh', st_file=True)
                Path('temp.smesh').unlink()
                st.session_state[f'{self.key}'] = vtk_mesh
                st.session_state[f'{self.key}_done'] = True


from .BaseElement import BaseElement
from .BaseFace import BaseFace
from .TriangleFace import TriangleFace
from .QuadrilateralFace import QuadrilateralFace
from .HexahedraElement import HexahedraElement
from .TetrahedraElement import TetrahedraElement
from .WedgeElement import WedgeElement
